from typing import Any, Dict

from pydantic import Field, field_validator

from luxonis_ml.data.utils import ImageType, MediaType
from luxonis_ml.utils import BaseModelExtraForbid


class LuxonisComponent(BaseModelExtraForbid):
    """Abstraction for a piece of media within a source. Most commonly,

    this abstracts an image sensor.
    @type name: str
    @ivar name: A recognizable name for the component.
    @type media_type: L{MediaType}
    @ivar media_type: Enum for the type of media for the component.
    @type image_type: L{ImageType}
    @ivar image_type: Enum for the image type. Only used if
        C{media_type==MediaType.IMAGE}.
    """

    name: str
    media_type: MediaType = MediaType.IMAGE
    image_type: ImageType = ImageType.COLOR


class LuxonisSource(BaseModelExtraForbid):
    """Abstracts the structure of a dataset and which components/media
    are included.

    For example, with an U{OAK-D<https://docs.luxonis.com/projects/hardware
    /en/latest/pages/BW1098OAK/>}, you can have a source with 4 image
    components: C{rgb} (color), C{left} (mono), C{right} (mono), and C{depth}.

    @type name: str
    @ivar name: A recognizable name for the source. Defaults to "default".

    @type components: Optional[List[LuxonisComponent]]
    @ivar components: If not using the default configuration, a list of
    L{LuxonisComponent} to group together in the source.

    @type main_component: Optional[str]
    @ivar main_component: The name of the component that should be
        primarily visualized.
    """

    name: str = "default"
    components: Dict[str, LuxonisComponent] = Field(
        default_factory=lambda: {"default": LuxonisComponent(name="default")}
    )
    main_component: str = "default"

    def merge_with(self, other: "LuxonisSource") -> "LuxonisSource":
        """Merge two sources together.

        @type other: LuxonisSource
        @param other: The other source to merge with.
        @rtype: LuxonisSource
        @return: A new source with the components of both sources.
        """
        components = self.components.copy()
        components.update(other.components)
        return LuxonisSource(
            name=self.name,
            components=components,
            main_component=self.main_component,
        )

    @field_validator("components", mode="before")
    @classmethod
    def validate_components(cls, components: Any) -> Any:
        if isinstance(components, list):
            migrated_components = {}
            for component in components:
                component = LuxonisComponent(**component)
                migrated_components[component.name] = component
            return migrated_components
        return components



from typing import Dict, Iterator, List, Tuple, TypeVar

import numpy as np


def preprocess_mask(seg: np.ndarray) -> np.ndarray:
    return seg.transpose(1, 2, 0)


def preprocess_bboxes(bboxes: np.ndarray, bbox_counter: int) -> np.ndarray:
    bboxes = bboxes[:, [1, 2, 3, 4, 0]]

    # Adding 1e-6 to avoid zero width or height.
    bboxes[:, 2] += bboxes[:, 0] + 1e-6
    bboxes[:, 3] += bboxes[:, 1] + 1e-6

    # Used later to filter out instance tasks associated
    # with bboxes that were removed during augmentations.
    indices = np.arange(
        bbox_counter, bboxes.shape[0] + bbox_counter, dtype=bboxes.dtype
    )[:, None]
    return np.concatenate((bboxes, indices), axis=1)


def preprocess_keypoints(
    keypoints: np.ndarray, height: int, width: int
) -> np.ndarray:
    keypoints = np.reshape(keypoints, (-1, 3))
    keypoints[:, 0] *= width
    keypoints[:, 1] *= height
    return keypoints


def postprocess_mask(mask: np.ndarray) -> np.ndarray:
    if mask.ndim == 2:
        return mask[None, ...]

    return mask.transpose(2, 0, 1)


def postprocess_bboxes(bboxes: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    area_threshold = 0.0004  # 0.02 * 0.02 Small area threshold to remove invalid bboxes and respective keypoints.
    if bboxes.size == 0:
        return np.zeros((0, 5)), np.zeros((0,), dtype=np.uint8)
    ordering = bboxes[:, -1]
    raw_bboxes = bboxes[:, :-1]
    raw_bboxes[:, 2] -= raw_bboxes[:, 0]
    raw_bboxes[:, 3] -= raw_bboxes[:, 1]
    widths = raw_bboxes[:, 2]
    heights = raw_bboxes[:, 3]
    areas = widths * heights

    valid_mask = areas >= area_threshold
    raw_bboxes = raw_bboxes[valid_mask]
    refined_ordering = ordering[valid_mask]

    out_bboxes = raw_bboxes[:, [4, 0, 1, 2, 3]]

    return out_bboxes, refined_ordering.astype(int)


def postprocess_keypoints(
    keypoints: np.ndarray,
    bboxes_ordering: np.ndarray,
    image_height: int,
    image_width: int,
    n_keypoints: int,
) -> np.ndarray:
    keypoints = keypoints[:, : (n_keypoints * 3)]
    keypoints = keypoints.reshape(-1, n_keypoints, 3)

    keypoints = keypoints[bboxes_ordering]

    x = keypoints[..., 0]
    y = keypoints[..., 1]
    v = keypoints[..., 2]

    in_bounds = (x >= 0) & (x < image_width) & (y >= 0) & (y < image_height)

    v[~in_bounds] = 0

    x = np.clip(x, 0, image_width)
    y = np.clip(y, 0, image_height)

    x /= image_width
    y /= image_height

    keypoints[..., 0] = x
    keypoints[..., 1] = y
    keypoints[..., 2] = v

    return keypoints.reshape(-1, n_keypoints * 3)


T = TypeVar("T")


def yield_batches(
    data_batch: List[Dict[str, T]], batch_size: int
) -> Iterator[Dict[str, List[T]]]:
    """Yield batches of data.

    @type data_batch: List[Dict[str, Any]]
    @param data_batch: List of dictionaries containing data.
    @type batch_size: int
    @param batch_size: Size of the batch.
    @rtype: Iterator[Dict[str, List[Any]]]
    @return: Generator of batches of data.
    """
    for i in range(0, len(data_batch), batch_size):
        yield {
            target: [data[target] for data in data_batch[i : i + batch_size]]
            for target in data_batch[0]
        }



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import cv2
import numpy as np
import polars as pl

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class SegmentationMaskDirectoryParser(BaseParser):
    """Parses directory with SegmentationMask annotations to LDF.

    Expected format::

        dataset_dir/
        ├── train/
        │   ├── img1.jpg
        │   ├── img1_mask.png
        │   ├── ...
        │   └── _classes.csv
        ├── valid/
        └── test/

    C{_classes.csv} contains mappings between pixel value and class name.

    This is one of the formats that can be generated by
    U{Roboflow <https://roboflow.com/>}.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        if not split_path.exists():
            return None
        if not (split_path / "_classes.csv").exists():
            return None
        masks = list(split_path.glob("*_mask.*"))
        for mask_path in masks:
            img_path = split_path / f"{mask_path.stem[:-5]}.jpg"
            if not img_path.exists():
                return None
        return {
            "image_dir": split_path,
            "seg_dir": split_path,
            "classes_path": split_path / "_classes.csv",
        }

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        for split in ["train", "valid", "test"]:
            split_path = dataset_dir / split
            if (
                SegmentationMaskDirectoryParser.validate_split(split_path)
                is None
            ):
                return False
        return True

    def from_dir(
        self, dataset_dir: Path
    ) -> Tuple[List[Path], List[Path], List[Path]]:
        added_train_imgs = self._parse_split(
            image_dir=dataset_dir / "train",
            seg_dir=dataset_dir / "train",
            classes_path=dataset_dir / "train" / "_classes.csv",
        )
        added_val_imgs = self._parse_split(
            image_dir=dataset_dir / "valid",
            seg_dir=dataset_dir / "valid",
            classes_path=dataset_dir / "valid" / "_classes.csv",
        )
        added_test_imgs = self._parse_split(
            image_dir=dataset_dir / "test",
            seg_dir=dataset_dir / "test",
            classes_path=dataset_dir / "test" / "_classes.csv",
        )
        return added_train_imgs, added_val_imgs, added_test_imgs

    def from_split(
        self, image_dir: Path, seg_dir: Path, classes_path: Path
    ) -> ParserOutput:
        """Parses annotations with SegmentationMask format to LDF.

        Annotations include classification and segmentation.

        @type image_dir: Path
        @param image_dir: Path to directory with images
        @type seg_dir: Path
        @param seg_dir: Path to directory with segmentation mask
        @type classes_path: Path
        @param classes_path: Path to CSV file with class names
        @rtype: L{ParserOutput}
        @return: Annotation generator, list of classes names, skeleton
            dictionary for keypoints and list of added images
        """

        # NOTE: space prefix included
        idx_class = " Class"

        df = pl.read_csv(classes_path).filter(pl.col(idx_class).is_not_null())
        class_names = df[idx_class].to_list()

        def generator() -> DatasetIterator:
            for mask_path in seg_dir.glob("*_mask.*"):
                image_path = next(image_dir.glob(f"{mask_path.stem[:-5]}.*"))
                file = str(image_path.absolute().resolve())
                mask = cv2.imread(str(mask_path), cv2.IMREAD_GRAYSCALE)

                ids = np.unique(mask)
                for id in ids:
                    class_name = class_names[id]

                    curr_seg_mask = np.zeros_like(mask)
                    curr_seg_mask[mask == id] = 1
                    yield {
                        "file": file,
                        "annotation": {
                            "class": class_name,
                            "segmentation": {"mask": curr_seg_mask},
                        },
                    }

        added_images = self._get_added_images(generator())
        return generator(), {}, added_images



import tempfile
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import pytest
import yaml
from pydantic import BaseModel

from luxonis_ml.typing import Params
from luxonis_ml.utils import environ
from luxonis_ml.utils.config import LuxonisConfig

CONFIG_DATA = {
    "sub_config": {
        "str_sub_param": "sub_param",
        "int_sub_param": 42,
        "float_sub_param": 1.0,
        "list_sub_param": [],
        "dict_sub_param": {},
    },
    "sub_config_default": {
        "str_sub_param": "sub_param_default",
    },
}


@pytest.fixture
def config_file() -> str:
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(yaml.dump(CONFIG_DATA).encode())
    return f.name


class SubConfigDefault(BaseModel):
    str_sub_param: str = "sub_param"
    int_sub_param: int = 42
    float_sub_param: float = 1.0

    list_sub_param: List[int] = []
    dict_sub_param: Dict[str, int] = {}


class SubConfig(BaseModel):
    str_sub_param: str
    int_sub_param: int
    float_sub_param: float

    list_sub_param: List[int]
    dict_sub_param: Dict[str, int]


class ListConfig(BaseModel):
    int_list_param: int
    float_list_param: float = 1.0
    str_list_param: Optional[str] = None


class UnsafeConfig(BaseModel):
    tuple_param: Tuple[int, int] = (1, 2)
    path_param: Path = Path.cwd()


class Config(LuxonisConfig):
    sub_config: SubConfig
    sub_config_default: SubConfigDefault = SubConfigDefault()

    optional_int: Optional[int] = None
    list_config: List[ListConfig] = []
    nested_list_param: List[List[int]] = []
    nested_dict_param: Dict[str, Dict[str, int]] = {}
    unsafe_config: UnsafeConfig = UnsafeConfig()


def test_invalid_config_path():
    with pytest.raises(FileNotFoundError):
        Config.get_config("invalid_path")

    with pytest.raises(
        ValueError, match="At least one of `cfg` or `overrides`"
    ):
        Config.get_config(None)


def test_config_simple(config_file: str):
    cfg = Config.get_config(config_file)
    assert (
        cfg.sub_config.str_sub_param
        == CONFIG_DATA["sub_config"]["str_sub_param"]
    )
    assert (
        cfg.sub_config.int_sub_param
        == CONFIG_DATA["sub_config"]["int_sub_param"]
    )
    assert (
        cfg.sub_config.float_sub_param
        == CONFIG_DATA["sub_config"]["float_sub_param"]
    )


def test_config_simple_override(config_file: str):
    overrides: Params = {
        "sub_config.str_sub_param": "sub_param_override",
    }
    cfg = Config.get_config(config_file, overrides)
    assert (
        cfg.sub_config.str_sub_param == overrides["sub_config.str_sub_param"]
    )
    assert (
        cfg.sub_config.int_sub_param
        == CONFIG_DATA["sub_config"]["int_sub_param"]
    )
    assert (
        cfg.sub_config.float_sub_param
        == CONFIG_DATA["sub_config"]["float_sub_param"]
    )


def test_config_list_override(config_file: str):
    overrides = ["sub_config.str_sub_param", "sub_param_override"]
    cfg = Config.get_config(config_file, overrides)
    assert cfg.sub_config.str_sub_param == overrides[1]
    assert (
        cfg.sub_config.int_sub_param
        == CONFIG_DATA["sub_config"]["int_sub_param"]
    )
    assert (
        cfg.sub_config.float_sub_param
        == CONFIG_DATA["sub_config"]["float_sub_param"]
    )
    with pytest.raises(ValueError, match="length is not divisible by 2"):
        Config.get_config(config_file, ["sub_config.str_sub_param"])


def test_config_nested_structure(config_file: str):
    cfg = Config.get_config(config_file)
    # Asserting nested configurations
    assert isinstance(cfg.sub_config.dict_sub_param, dict)
    assert isinstance(cfg.sub_config.list_sub_param, list)
    assert cfg.sub_config_default.str_sub_param == "sub_param_default"
    assert cfg.sub_config_default.int_sub_param == 42


def test_config_optional_params(config_file: str):
    cfg = Config.get_config(config_file)
    # Testing optional parameters
    assert cfg.optional_int is None
    assert isinstance(cfg.list_config, list)
    assert len(cfg.list_config) == 0


def test_config_override_list(config_file: str):
    overrides = {
        "list_config.0.int_list_param": 10,
        "list_config.0.float_list_param": 2.5,
        "list_config.0.str_list_param": "test",
        "list_config.1.int_list_param": 20,
        "nested_list_param.0": [30],
        "nested_list_param.0.1": 40,
    }
    cfg = Config.get_config(config_file, overrides)
    # Testing list configurations
    assert len(cfg.list_config) == 2
    assert cfg.list_config[0].int_list_param == 10
    assert cfg.list_config[0].float_list_param == 2.5
    assert cfg.list_config[0].str_list_param == "test"
    assert cfg.list_config[1].int_list_param == 20
    assert cfg.list_config[1].float_list_param == 1.0
    assert cfg.list_config[1].str_list_param is None
    assert cfg.nested_list_param[0][0] == 30
    assert cfg.nested_list_param[0][1] == 40


def test_config_list_override_json(config_file: str):
    overrides = {
        "list_config": '[{"int_list_param": 10, "float_list_param": 2.5, '
        '"str_list_param": "test"}, {"int_list_param": 20} ]',
        "list_config.2.int_list_param": 30,
    }

    cfg = Config.get_config(config_file, overrides)

    assert len(cfg.list_config) == 3
    assert cfg.list_config[0].int_list_param == 10
    assert cfg.list_config[0].float_list_param == 2.5
    assert cfg.list_config[0].str_list_param == "test"
    assert cfg.list_config[1].int_list_param == 20
    assert cfg.list_config[1].float_list_param == 1.0
    assert cfg.list_config[1].str_list_param is None
    assert cfg.list_config[2].int_list_param == 30


def test_invalid_config(config_file: str):
    overrides = {
        "list_config.-1.int_list_param": 10,
        "list_config.0.float_list_param": [],
        "list_config.0.str_list_param": "test",
        "sub_config.str_sub_param": 10,
        "list_config.2": 30,
        "sub_config.5": 30,
        "5.sub_config": 30,
        "nested_list_param.non": [30],
        "nested_list_param.0.non": 40,
        "sub_config.str_sub_param.non": "sub_param_override",
        "list_config": '[{"int_list_param": 10, "float_list_param": 2.5, '
        '"str_list_param": "test"}, {"int_list_param": 20.5} ]',
        "extra_param": "test",
    }
    for key, value in overrides.items():
        with pytest.raises(ValueError):  # noqa: PT011
            Config.get_config(config_file, {key: value})


def test_from_dict():
    config_dict = deepcopy(CONFIG_DATA)
    config_dict["nested_list_param"] = [[1, 2]]
    config_dict["nested_dict_param"] = {"a": {"b": 1}}
    cfg = Config.get_config(
        config_dict,
        {
            "nested_list_param.0.1": 3,
            "nested_dict_param.a.b": 2,
        },
    )
    assert (
        cfg.sub_config.str_sub_param
        == CONFIG_DATA["sub_config"]["str_sub_param"]
    )
    assert (
        cfg.sub_config.int_sub_param
        == CONFIG_DATA["sub_config"]["int_sub_param"]
    )
    assert (
        cfg.sub_config.float_sub_param
        == CONFIG_DATA["sub_config"]["float_sub_param"]
    )
    assert cfg.nested_list_param[0][1] == 3
    assert cfg.nested_dict_param["a"]["b"] == 2


def test_save(config_file: str):
    cfg = Config.get_config(config_file)
    with tempfile.NamedTemporaryFile(delete=False) as f:
        cfg.save_data(f.name)
    cfg2 = Config.get_config(f.name)
    assert cfg.__repr__() == cfg2.__repr__()
    assert cfg.get_json_schema() == cfg2.get_json_schema()


def test_get(config_file: str):
    cfg = Config.get_config(
        config_file,
        {"nested_list_param": [[1, 2]], "nested_dict_param": {"a": {"b": 1}}},
    )
    assert (
        cfg.get("sub_config.str_sub_param")
        == CONFIG_DATA["sub_config"]["str_sub_param"]
    )
    assert (
        cfg.get("sub_config.int_sub_param")
        == CONFIG_DATA["sub_config"]["int_sub_param"]
    )
    assert cfg.get("list_config.2.int_list_param", -56) == -56
    assert cfg.get("list_config.0.int_list_param") is None
    assert cfg.get("nested_list_param.0.1") == 2
    assert cfg.get("nested_dict_param.a.b") == 1
    assert cfg.get("nested_dict_param.a.c") is None

    assert cfg.get("sub_config.str_sub_param.non", "default") == "default"

    with pytest.raises(ValueError, match="Can't access list with non-int key"):
        cfg.get("list_config.index.int_list_param")


def test_environ():
    assert environ.model_dump() == {}


def test_safe_load(config_file: str):
    cfg = Config.get_config(config_file)
    with tempfile.NamedTemporaryFile(delete=False) as f:
        cfg.save_data(f.name)

    cfg2 = Config.get_config(f.name)
    assert cfg == cfg2



from enum import Enum


class MediaType(Enum):
    """Individual file type."""

    IMAGE = "image"
    VIDEO = "video"
    POINTCLOUD = "pointcloud"


class ImageType(Enum):
    """Image type for IMAGE HType."""

    COLOR = "color"  # 3 channel BGR or RGB (uint8)
    MONO = "mono"  # 1 channel (uint8)
    DISPARITY = "disparity"  # disparity or depth (uint16)


class BucketType(Enum):
    """Whether bucket storage is internal to Luxonis or not."""

    INTERNAL = "internal"
    EXTERNAL = "external"


class BucketStorage(Enum):
    """Underlying object storage for a bucket."""

    LOCAL = "local"
    S3 = "s3"
    GCS = "gcs"
    AZURE_BLOB = "azure"


class UpdateMode(Enum):
    """Update mode for the dataset."""

    ALL = "all"
    MISSING = "missing"



from pathlib import Path
from typing import Optional, Set, TypedDict

import polars as pl

from luxonis_ml.typing import PathType


class ParquetRecord(TypedDict):
    file: str
    source_name: str
    task_name: str
    class_name: Optional[str]
    instance_id: Optional[int]
    task_type: Optional[str]
    annotation: Optional[str]


class ParquetFileManager:
    def __init__(self, directory: PathType, num_rows: int = 100_000) -> None:
        """Manages the insertion of data into parquet files.

        @type directory: str
        @param directory: The local directory in which parquet files are
            stored.
        @type num_rows: int
        @param num_rows: The maximum number of rows permitted in a
            parquet file before another file is created.
        """

        self.dir = Path(directory)
        self.parquet_files = list(self.dir.glob("*.parquet"))
        self.num_rows = num_rows

        self.num = self._find_num() if self.parquet_files else 0
        self.current_file = self._generate_filename(self.num)

        self._read()

    def _find_num(self) -> int:
        return max(int(file.stem) for file in self.parquet_files)

    def _generate_filename(self, num: int) -> Path:
        return self.dir / f"{num:010d}.parquet"

    def _read(self) -> None:
        if self.current_file.exists():
            df = pl.read_parquet(self.current_file)
            self.buffer = df.to_dict(as_series=False)
            self.row_count = len(df)
        else:
            self.row_count = 0
            self.buffer = {}

    def _initialize_data(self, data: ParquetRecord) -> None:
        for key in data:
            self.buffer[key] = []
        self.buffer["uuid"] = []

    def write(self, uuid: str, data: ParquetRecord) -> None:
        """Writes a row to the current working parquet file.

        @type data: Dict
        @param data: A dictionary representing annotations, mapping
            annotation types to values.
        """

        if not self.buffer:
            self._initialize_data(data)

        for key in data:
            self.buffer[key].append(data[key])

        self.buffer["uuid"].append(uuid)

        self.row_count += 1
        if self.row_count % self.num_rows == 0:
            self._flush()
            self.num += 1
            self.current_file = self._generate_filename(self.num)
            self._read()

    def remove_duplicate_uuids(self, overwrite_uuids: Set[str]) -> None:
        self._flush()

        for parquet_file in self.dir.glob("*.parquet"):
            if parquet_file.is_file():
                df = pl.read_parquet(parquet_file)
                df = df.filter(~pl.col("uuid").is_in(list(overwrite_uuids)))
                parquet_file.unlink()
                df.write_parquet(parquet_file)

        self._read()

    def _flush(self) -> None:
        """Writes buffered data to parquet."""

        if self.buffer:
            df = pl.DataFrame(self.buffer)
            df.write_parquet(self.current_file)

    def __enter__(self) -> "ParquetFileManager":
        return self

    def __exit__(self, *_) -> None:
        self._flush()



from abc import ABC
from typing import List, Optional, Union

from pydantic import BaseModel, Field

from .head_metadata import (
    HeadClassificationMetadata,
    HeadMetadata,
    HeadObjectDetectionMetadata,
    HeadObjectDetectionSSDMetadata,
    HeadSegmentationMetadata,
    HeadYOLOMetadata,
)


class Head(BaseModel, ABC):
    """Represents head of a model.

    @type name: str | None
    @ivar name: Optional name of the head.
    @type parser: str
    @ivar parser: Name of the parser responsible for processing the
        models output.
    @type outputs: List[str] | None
    @ivar outputs: Specify which outputs are fed into the parser. If
        None, all outputs are fed.
    @type metadata: C{HeadMetadata} | C{HeadObjectDetectionMetadata} |
        C{HeadClassificationMetadata} |
        C{HeadObjectDetectionSSDMetadata} | C{HeadSegmentationMetadata}
        | C{HeadYOLOMetadata}
    @ivar metadata: Metadata of the parser.
    """

    name: Optional[str] = Field(None, description="Optional name of the head.")
    parser: str = Field(
        description="Name of the parser responsible for processing the models output."
    )
    metadata: Union[
        HeadObjectDetectionMetadata,
        HeadClassificationMetadata,
        HeadObjectDetectionSSDMetadata,
        HeadSegmentationMetadata,
        HeadYOLOMetadata,
        HeadMetadata,
    ] = Field(description="Metadata of the parser.")
    outputs: Optional[List[str]] = Field(
        None,
        description="Specify which outputs are fed into the parser. If None, all outputs are fed.",
    )


HeadType = Head



import pytest

from luxonis_ml.data import Metadata
from luxonis_ml.data.datasets.source import LuxonisSource


@pytest.fixture
def basic_metadata() -> Metadata:
    return Metadata(
        source=None,
        ldf_version="2.0.0",
        classes={"task1": {"class1": 0, "class2": 1}},
        tasks={"task1": ["subtask1", "subtask2"]},
        skeletons={"task1": {"labels": ["head", "tail"], "edges": [(0, 1)]}},
        categorical_encodings={"cat1": {"a": 0, "b": 1}},
        metadata_types={"field1": "str", "field2": "int"},
    )


def test_merge_with_different_versions():
    """Test that merging metadata with different LDF versions raises
    ValueError."""
    metadata1 = Metadata(
        ldf_version="2.0.0",
        source=None,
        classes={},
        tasks={},
        skeletons={},
        categorical_encodings={},
        metadata_types={},
    )
    metadata2 = Metadata(
        ldf_version="2.0",
        source=None,
        classes={},
        tasks={},
        skeletons={},
        categorical_encodings={},
        metadata_types={},
    )

    with pytest.raises(
        ValueError, match="Cannot merge metadata with different LDF versions"
    ):
        metadata1.merge_with(metadata2)


def test_merge_classes(basic_metadata: Metadata):
    """Test merging of classes dictionaries."""
    other_metadata = Metadata(
        source=None,
        ldf_version="2.0.0",
        classes={"task1": {"class2": 1, "class3": 2}, "task2": {"classA": 0}},
        tasks={},
        skeletons={},
        categorical_encodings={},
        metadata_types={},
    )

    merged = basic_metadata.merge_with(other_metadata)

    assert "task1" in merged.classes
    assert "task2" in merged.classes
    assert len(merged.classes["task1"]) == 3
    assert all(
        c in merged.classes["task1"] for c in ["class1", "class2", "class3"]
    )
    assert merged.classes["task2"] == {"classA": 0}


def test_merge_tasks(basic_metadata: Metadata):
    """Test merging of tasks dictionaries."""
    other_metadata = Metadata(
        source=None,
        ldf_version="2.0.0",
        classes={},
        tasks={"task1": ["subtask2", "subtask3"], "task2": ["other_subtask"]},
        skeletons={},
        categorical_encodings={},
        metadata_types={},
    )

    merged = basic_metadata.merge_with(other_metadata)

    assert set(merged.tasks["task1"]) == {"subtask1", "subtask2", "subtask3"}
    assert merged.tasks["task2"] == ["other_subtask"]


def test_merge_categorical_encodings(basic_metadata: Metadata):
    """Test merging of categorical encodings."""
    other_metadata = Metadata(
        source=None,
        ldf_version="2.0.0",
        classes={},
        tasks={},
        skeletons={},
        categorical_encodings={"cat1": {"b": 1, "c": 2}, "cat2": {"x": 0}},
        metadata_types={},
    )

    merged = basic_metadata.merge_with(other_metadata)

    assert "cat1" in merged.categorical_encodings
    assert "cat2" in merged.categorical_encodings
    assert merged.categorical_encodings["cat1"]["c"] == 2
    assert merged.categorical_encodings["cat2"] == {"x": 0}


def test_merge_sources():
    """Test merging of sources."""
    source1 = LuxonisSource()
    source2 = LuxonisSource()

    metadata1 = Metadata(
        source=source1,
        ldf_version="2.0.0",
        classes={},
        tasks={},
        skeletons={},
        categorical_encodings={},
        metadata_types={},
    )

    metadata2 = Metadata(
        source=source2,
        ldf_version="2.0.0",
        classes={},
        tasks={},
        skeletons={},
        categorical_encodings={},
        metadata_types={},
    )

    merged = metadata1.merge_with(metadata2)
    assert merged.source is not None


def test_merge_with_none_sources():
    """Test merging when one or both sources are None."""
    source = LuxonisSource()

    metadata1 = Metadata(
        source=None,
        ldf_version="2.0.0",
        classes={},
        tasks={},
        skeletons={},
        categorical_encodings={},
        metadata_types={},
    )

    metadata2 = Metadata(
        source=source,
        ldf_version="2.0.0",
        classes={},
        tasks={},
        skeletons={},
        categorical_encodings={},
        metadata_types={},
    )

    merged = metadata1.merge_with(metadata2)
    assert merged.source == source

    merged = metadata2.merge_with(metadata1)
    assert merged.source == source


def test_merge_metadata_types(basic_metadata: Metadata):
    """Test merging of metadata types."""
    other_metadata = Metadata(
        source=None,
        ldf_version="2.0.0",
        classes={},
        tasks={},
        skeletons={},
        categorical_encodings={},
        metadata_types={"field2": "int", "field3": "float"},
    )

    merged = basic_metadata.merge_with(other_metadata)

    assert merged.metadata_types["field1"] == "str"
    assert merged.metadata_types["field2"] == "int"
    assert merged.metadata_types["field3"] == "float"



from typing import List

from luxonis_ml.typing import all_not_none, any_not_none, check_type


def test_all_not_none():
    assert all_not_none([1, 2, 3])
    assert not all_not_none([1, 2, None])
    assert not all_not_none([None, None, None])
    assert all_not_none([])


def test_any_not_none():
    assert any_not_none([1, 2, 3])
    assert any_not_none([1, 2, None])
    assert not any_not_none([None, None, None])
    assert not any_not_none([])


def test_check_type():
    assert check_type(1, int)
    assert not check_type(1, str)
    assert check_type([1, 2, 3], List[int])
    assert not check_type([1, 2, 3], List[str])



from .duplicate import find_similar
from .mistakes import find_mismatches_centroids, find_mismatches_knn
from .OOD import isolation_forest_OOD, leverage_OOD
from .representative import (
    calculate_similarity_matrix,
    find_representative_greedy,
    find_representative_kmedoids,
)

__all__ = [
    "calculate_similarity_matrix",
    "find_mismatches_centroids",
    "find_mismatches_knn",
    "find_representative_greedy",
    "find_representative_kmedoids",
    "find_similar",
    "isolation_forest_OOD",
    "leverage_OOD",
]



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class ClassificationDirectoryParser(BaseParser):
    """Parses directory with ClassificationDirectory annotations to LDF.

    Expected format::

        dataset_dir/
        ├── train/
        │   ├── class1/
        │   │   ├── img1.jpg
        │   │   ├── img2.jpg
        │   │   └── ...
        │   ├── class2/
        │   └── ...
        ├── valid/
        └── test/

    This is one of the formats that can be generated by
    U{Roboflow <https://roboflow.com/>}.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        if not split_path.exists():
            return None
        classes = [
            d
            for d in split_path.iterdir()
            if d.is_dir() and d.name not in {"train", "valid", "test"}
        ]
        if not classes:
            return None
        fnames = [f for f in split_path.iterdir() if f.is_file()]
        if fnames:
            return None
        return {"class_dir": split_path}

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        for split in ["train", "valid", "test"]:
            split_path = dataset_dir / split
            if (
                ClassificationDirectoryParser.validate_split(split_path)
                is None
            ):
                return False
        return True

    def from_dir(
        self, dataset_dir: Path
    ) -> Tuple[List[Path], List[Path], List[Path]]:
        added_train_imgs = self._parse_split(class_dir=dataset_dir / "train")
        added_val_imgs = self._parse_split(class_dir=dataset_dir / "valid")
        added_test_imgs = self._parse_split(class_dir=dataset_dir / "test")
        return added_train_imgs, added_val_imgs, added_test_imgs

    def from_split(self, class_dir: Path) -> ParserOutput:
        """Parses annotations from classification directory format to
        LDF. Annotations include classification.

        @type class_dir: Path
        @param class_dir: Path to top level directory
        @rtype: L{ParserOutput}
        @return: Annotation generator, list of classes names, skeleton
            dictionary for keypoints and list of added images.
        """
        class_names = [d.name for d in class_dir.iterdir() if d.is_dir()]

        def generator() -> DatasetIterator:
            for class_name in class_names:
                for img_path in (class_dir / class_name).iterdir():
                    yield {
                        "file": str(img_path.absolute().resolve()),
                        "annotation": {"class": class_name},
                    }

        added_images = self._get_added_images(generator())

        return generator(), {}, added_images



import json
from collections import defaultdict
from typing import Any, Dict, Iterator, List, Optional, Tuple

import numpy as np
import polars as pl
from loguru import logger
from pycocotools import mask as mask_utils

from luxonis_ml.data.utils.task_utils import task_is_metadata
from luxonis_ml.typing import RGB


def rgb_to_bool_masks(
    segmentation_mask: np.ndarray,
    class_colors: Dict[str, RGB],
    add_background_class: bool = False,
) -> Iterator[Tuple[str, np.ndarray]]:
    """Helper function to convert an RGB segmentation mask to boolean
    masks for each class.

    Example:

        >>> segmentation_mask = np.array([
        ...     [[0, 0, 0], [255, 0, 0], [0, 255, 0]],
        ...     [[0, 0, 0], [0, 255, 0], [0, 0, 255]],
        ...     ], dtype=np.uint8)
        >>> class_colors = {
        ...     "red": (255, 0, 0),
        ...     "green": (0, 255, 0),
        ...     "blue": (0, 0, 255),
        ... }
        >>> for class_name, mask in rgb_to_bool_masks(
        ...     segmentation_mask,
        ...     class_colors,
        ...     add_background_class=True,
        ... ):
        ...     print(class_name, np.array2string(mask, separator=", "))
        background [[ True, False, False],
                    [ True, False, False]]
        red        [[False,  True, False],
                    [False, False, False]]
        green      [[False, False,  True],
                    [False, True, False]]
        blue       [[False, False, False],
                    [False, False,  True]]

    @type segmentation_mask: npt.NDArray[np.uint8]
    @param segmentation_mask: An RGB segmentation mask where each pixel
        is colored according to the class it belongs to.
    @type class_colors: Dict[str, Tuple[int, int, int]]
    @param class_colors: A dictionary mapping class names to RGB colors.
    @type add_background_class: bool
    @param add_background_class: Whether to add a background class with a mask for all pixels
        that do not belong to any class. The class name will be set to "background".
        The background class will be yielded first. Default is False.
    @rtype: Iterator[Tuple[str, npt.NDArray[np.bool_]]]
    @return: An iterator of tuples where the first element is the class name and
        the second element is a boolean mask for that class.
    """
    color_to_id = {
        tuple(color): i for i, color in enumerate(class_colors.values())
    }

    lookup_table = np.zeros((256, 256, 256), dtype=np.uint8)
    for color, id in color_to_id.items():
        lookup_table[color[0], color[1], color[2]] = id + 1

    segmentation_ids = lookup_table[
        segmentation_mask[:, :, 0],
        segmentation_mask[:, :, 1],
        segmentation_mask[:, :, 2],
    ]

    if add_background_class:
        background_mask = segmentation_ids == 0
        yield "background", background_mask

    for class_name, color in class_colors.items():
        class_id = color_to_id[tuple(color)] + 1
        yield class_name, segmentation_ids == class_id


def infer_task(
    old_task: str,
    class_name: Optional[str],
    current_classes: Dict[str, Dict[str, int]],
) -> str:
    if not hasattr(infer_task, "_logged_infered_classes"):
        infer_task._logged_infered_classes = defaultdict(bool)

    def _log_once(
        cls_: Optional[str], task: str, message: str, level: str = "info"
    ) -> None:
        if not infer_task._logged_infered_classes[(cls_, task)]:
            infer_task._logged_infered_classes[(cls_, task)] = True
            getattr(logger, level)(message)

    infered_task = None

    for task, classes in current_classes.items():
        if class_name in classes:
            if infered_task is not None:
                _log_once(
                    class_name,
                    infered_task,
                    f"Class '{class_name}' is ambiguous between "
                    f"tasks '{infered_task}' and '{task}'. "
                    "Task inference failed.",
                    "warning",
                )
                infered_task = None
                break
            infered_task = task
    if infered_task is None:
        _log_once(
            class_name,
            old_task,
            f"Class '{class_name}' doesn't belong to any existing task. "
            f"Autogenerated task '{old_task}' will be used.",
            "info",
        )
    else:
        _log_once(
            class_name,
            infered_task,
            f"Class '{class_name}' infered to belong to task '{infered_task}'",
        )
        return infered_task

    return old_task


def find_duplicates(df: pl.LazyFrame) -> Dict[str, List[Dict[str, Any]]]:
    """Collects information about duplicate UUIDs and duplicate
    annotations in the dataset.

    @type df: pl.LazyFrame
    @param df: Polars lazy frame containing dataset information.
    @rtype: Dict[str, List[Dict[str, Any]]]
    @return: A dictionary with two keys:
        - "duplicate_uuids": list of dicts with "uuid" as key and "files" as value
        - "duplicate_annotations": list of dicts with "file_name", "task_type",
          "task_name", "annotation", and "count"
    """

    result = {
        "duplicate_uuids": [],
        "duplicate_annotations": [],
    }

    # Find duplicate UUIDs
    uuid_file_pairs = (
        df.select(
            pl.col("uuid").map_elements(lambda x: x, return_dtype=pl.Utf8),
            pl.col("file").map_elements(lambda x: x, return_dtype=pl.Utf8),
        )
        .unique()
        .collect()
    )

    duplicates = (
        uuid_file_pairs.group_by("uuid")
        .agg(pl.col("file").n_unique().alias("file_count"))
        .filter(pl.col("file_count") > 1)
    )

    if not duplicates.is_empty():
        logger.warning("Found duplicate UUIDs in the dataset:")
        for row in duplicates.iter_rows():
            uuid = row[0]
            files = uuid_file_pairs.filter(pl.col("uuid") == uuid)[
                "file"
            ].to_list()
            result["duplicate_uuids"].append(
                {
                    "uuid": uuid,
                    "files": files,
                }
            )

    # Find duplicate annotations
    def is_all_zero_keypoints(annotation_str: str) -> bool:
        """Check if a keypoints annotation has all zeros (x, y,
        visibility)."""
        try:
            annotation = json.loads(annotation_str)
            if "keypoints" in annotation:
                for kp in annotation["keypoints"]:
                    if len(kp) >= 3:
                        x, y, v = kp[0], kp[1], kp[2]
                        if x != 0 or y != 0 or v != 0:
                            return False
                return True
        except (TypeError, ValueError):
            return False
        return False

    filtered_df = df.filter(
        ~(
            (pl.col("task_type") == "keypoints")
            & (
                pl.col("annotation").map_elements(
                    is_all_zero_keypoints, return_dtype=pl.Boolean
                )
            )
        )
    )
    duplicate_annotation = (
        filtered_df.group_by(
            "original_filepath",
            "task_type",
            "task_name",
            "annotation",
        )
        .agg(pl.len().alias("count"))
        .filter(pl.col("count") > 1)
        .filter(pl.col("annotation") != "{}")
    ).collect()

    for (
        file_name,
        task_type,
        task_name,
        annotation,
        count,
    ) in duplicate_annotation.iter_rows():
        if task_type == "segmentation":
            annotation = "<binary mask>"
        if not task_is_metadata(task_type):
            result["duplicate_annotations"].append(
                {
                    "file_name": file_name,
                    "task_type": task_type,
                    "task_name": task_name,
                    "annotation": annotation,
                    "count": count,
                }
            )

    return result


def warn_on_duplicates(df: pl.LazyFrame) -> None:
    """Logs warnings for duplicate UUIDs and annotations in the dataset.

    @type df: pl.LazyFrame
    @param df: Polars lazy frame containing dataset information.
    """
    duplicates_info = find_duplicates(df)
    if duplicates_info["duplicate_uuids"]:
        for item in duplicates_info["duplicate_uuids"]:
            logger.warning(
                f"UUID {item['uuid']} is the same for multiple files: {item['files']}"
            )

    for item in duplicates_info["duplicate_annotations"]:
        logger.warning(
            f"File '{item['file_name']}' of task '{item['task_name']}' has the "
            f"same '{item['task_type']}' annotation '{item['annotation']}' repeated {item['count']} times."
        )


def get_class_distributions(
    df: pl.LazyFrame,
) -> Dict[str, Dict[str, List[Dict[str, Any]]]]:
    """Gets class distribution info for non-classification tasks.

    @type df: pl.LazyFrame
    @param df: Polars lazy frame containing dataset information.
    @rtype: Dict[str, Dict[str, List[Dict[str, Any]]]]
    @return: A dictionary with task names as keys, and dictionaries with
        task types as keys and lists of dictionaries with class names
        and counts as values.
    """
    class_distribution_raw = (
        df.filter(pl.col("task_type") != "classification")
        .group_by(["task_name", "task_type", "class_name"])
        .agg(pl.count().alias("count"))
        .sort(["task_name", "task_type", "count"], descending=True)
        .collect()
        .to_dicts()
    )
    class_distributions: Dict[str, Dict[str, List[Dict[str, Any]]]] = {}
    for record in class_distribution_raw:
        task_name = record["task_name"]
        task_type = record["task_type"]
        class_distributions.setdefault(task_name, {}).setdefault(
            task_type, []
        ).append(
            {
                "class_name": record["class_name"],
                "count": record["count"],
            }
        )
    return class_distributions


def get_missing_annotations(df: pl.LazyFrame) -> List[str]:
    """Returns file paths that exist but have no annotations.

    @type df: pl.LazyFrame
    @param df: Polars lazy frame containing dataset information.
    @rtype: List[str]
    @return: A list of file paths that exist but have no annotations.
    """
    all_files = df.select(pl.col("file").unique())
    annotated_files = df.filter(
        pl.col("annotation").is_not_null() & (pl.col("annotation") != "")
    ).select(pl.col("file").unique())
    missing_files_df = all_files.join(
        annotated_files, on="file", how="anti"
    ).collect()
    return missing_files_df["file"].to_list()


def get_duplicates_info(df: pl.LazyFrame) -> Dict[str, Any]:
    """Collects and returns information about duplicate UUIDs and
    annotations.

    @type df: pl.DataFrame
    @param df: Polars DataFrame containing dataset information.
    @rtype: Dict[str, Any]
    @return: A dictionary with two keys:
        - "duplicate_uuids": list of dicts with "uuid" as key and "files" as value
        - "duplicate_annotations": list of dicts with "file_name", "task_name",
          "task_type", "annotation", and "count"
    """
    duplicates_info = find_duplicates(df)
    return {
        "duplicate_uuids": [
            {"uuid": item["uuid"], "files": item["files"]}
            for item in duplicates_info["duplicate_uuids"]
        ],
        "duplicate_annotations": [
            {
                "file_name": item["file_name"],
                "task_name": item["task_name"],
                "task_type": item["task_type"],
                "annotation": item["annotation"],
                "count": item["count"],
            }
            for item in duplicates_info["duplicate_annotations"]
        ],
    }


def get_heatmaps(
    df: pl.LazyFrame,
    sample_size: Optional[int] = None,
    downsample_factor: int = 5,
) -> Dict[str, Dict[str, List[List[int]]]]:
    """Generates heatmaps for bounding boxes, keypoints, and
    segmentations.

    @type df: pl.LazyFrame
    @param df: Polars lazy frame containing dataset information.
    @type sample_size: Optional[int]
    @param sample_size: Number of samples to take from the dataset.
        Default is None.
    @type downsample_factor: int
    @param downsample_factor: Factor to downsample the segmentation
        masks.
    @rtype: Dict[str, Dict[str, List[List[int]]]]
    @return: A dictionary with task names as keys, and dictionaries with
        task types as keys and lists of lists of integers representing
        the heatmaps as values
    """
    task_types = [
        "boundingbox",
        "keypoints",
        "segmentation",
        "instance_segmentation",
    ]
    grid_size = 15
    x_edges = np.linspace(0, 1, grid_size + 1)
    y_edges = np.linspace(0, 1, grid_size + 1)
    heatmaps: Dict[str, Dict[str, np.ndarray]] = {}

    annotations_df = df.filter(
        pl.col("task_type").is_not_null() & pl.col("task_name").is_not_null()
    )

    annotations_df = annotations_df.select(
        "task_name", "task_type", "annotation"
    ).collect()

    if sample_size is not None and sample_size > 0:
        if sample_size > len(annotations_df):
            sample_size = None
        else:
            annotations_df = annotations_df.sample(n=sample_size, shuffle=True)

    rows = annotations_df.iter_rows(named=True)
    for row in rows:
        task_name = row["task_name"]
        task_type = row["task_type"]
        annotation_str = row["annotation"]

        if task_type not in task_types or not annotation_str:
            continue

        try:
            annotation = json.loads(annotation_str)
        except (json.JSONDecodeError, TypeError):
            continue

        try:
            if task_type == "boundingbox":
                w, h = annotation["w"], annotation["h"]
                x = np.array([annotation["x"] + w / 2])
                y = np.array([annotation["y"] + h / 2])
            elif task_type == "keypoints":
                kps = [
                    kp
                    for kp in annotation.get("keypoints", [])
                    if len(kp) >= 3 and kp[2] > 0
                ]
                if not kps:
                    continue
                coords = np.array([(kp[0], kp[1]) for kp in kps])
                x = coords[:, 0]
                y = coords[:, 1]
            elif task_type in ["segmentation", "instance_segmentation"]:
                rle = {
                    "counts": annotation["counts"],
                    "size": [annotation["height"], annotation["width"]],
                }
                mask = mask_utils.decode(rle)  # type: ignore
                if downsample_factor > 1:
                    mask = mask[::downsample_factor, ::downsample_factor]
                rows_idx, cols_idx = np.where(mask)
                if rows_idx.size == 0:
                    continue
                h_, w_ = mask.shape
                x = cols_idx / w_
                y = rows_idx / h_
            else:
                continue

            xi = np.digitize(x, x_edges) - 1
            xi = np.clip(xi, 0, grid_size - 1)
            yi = np.digitize(y, y_edges) - 1
            yi = np.clip(yi, 0, grid_size - 1)
            indices = xi * grid_size + yi
            counts = np.bincount(
                indices, minlength=grid_size * grid_size
            ).reshape(grid_size, grid_size)

            if task_name not in heatmaps:
                heatmaps[task_name] = {}
            if task_type not in heatmaps[task_name]:
                heatmaps[task_name][task_type] = np.zeros(
                    (grid_size, grid_size), dtype=np.int64
                )
            heatmaps[task_name][task_type] += counts

        except KeyError as e:
            logger.warning(f"Missing key in annotation: {e}")
            continue

    result: Dict[str, Dict[str, List[List[int]]]] = {}
    for t_name, tasks in heatmaps.items():
        result[t_name] = {}
        for t_type, grid in tasks.items():
            result[t_name][t_type] = grid.T.tolist()

    return result



from .base_parser import BaseParser
from .classification_directory_parser import ClassificationDirectoryParser
from .coco_parser import COCOParser
from .create_ml_parser import CreateMLParser
from .darknet_parser import DarknetParser
from .luxonis_parser import LuxonisParser
from .segmentation_mask_directory_parser import SegmentationMaskDirectoryParser
from .solo_parser import SOLOParser
from .tensorflow_csv_parser import TensorflowCSVParser
from .voc_parser import VOCParser
from .yolov4_parser import YoloV4Parser
from .yolov6_parser import YoloV6Parser

__all__ = [
    "BaseParser",
    "COCOParser",
    "ClassificationDirectoryParser",
    "CreateMLParser",
    "DarknetParser",
    "LuxonisParser",
    "SOLOParser",
    "SegmentationMaskDirectoryParser",
    "TensorflowCSVParser",
    "VOCParser",
    "YoloV4Parser",
    "YoloV6Parser",
]



from typing import Any, Dict, Tuple

import albumentations as A
import cv2
import numpy as np
from typing_extensions import override

from luxonis_ml.data.utils.visualizations import resolve_color
from luxonis_ml.typing import RGB, Color


class LetterboxResize(A.DualTransform):
    mask_fill_value: RGB

    def __init__(
        self,
        height: int,
        width: int,
        interpolation: int = cv2.INTER_LINEAR,
        image_fill_value: Color = "black",
        mask_fill_value: int = 0,
        p: float = 1.0,
    ):
        """Augmentation to apply letterbox resizing to images. Also
        transforms masks, bboxes and keypoints to correct shape.

        @type height: int
        @param height: Desired height of the output.
        @type width: int
        @param width: Desired width of the output.
        @type interpolation: int
        @param interpolation: cv2 flag to specify interpolation used
            when resizing. Defaults to C{cv2.INTER_LINEAR}.
        @type image_fill_value: int
        @param image_fill_value: Padding value for images. Defaults to
            "black".
        @type mask_fill_value: int
        @param mask_fill_value: Padding value for masks. Must be an
            integer representing the class label. Defaults to C{0}
            (background class).
        @type p: float
        @param p: Probability of applying the transform. Defaults to
            C{1.0}.
        """

        super().__init__(p=p)

        self.height = height
        self.width = width
        self.interpolation = interpolation
        self.image_fill_value = resolve_color(image_fill_value)
        self.mask_fill_value = resolve_color(mask_fill_value)

    @property
    @override
    def targets(self) -> Dict[str, Any]:
        targets = super().targets
        targets["instance_mask"] = self.apply_to_mask
        return targets

    @override
    def update_params(
        self, params: Dict[str, Any], **kwargs
    ) -> Dict[str, Any]:
        """Updates augmentation parameters with the necessary metadata.

        @param params: The existing augmentation parameters dictionary.
        @type params: Dict[str, Any]
        @param kwargs: Additional keyword arguments to add the
            parameters.
        @type kwargs: Any
        @return: Updated dictionary containing the merged parameters.
        @rtype: Dict[str, Any]
        """

        params = super().update_params(params, **kwargs)
        pad_top, pad_bottom, pad_left, pad_right = self.compute_padding(
            params["rows"], params["cols"], self.height, self.width
        )

        params.update(
            {
                "pad_top": pad_top,
                "pad_bottom": pad_bottom,
                "pad_left": pad_left,
                "pad_right": pad_right,
            }
        )

        return params

    @staticmethod
    def compute_padding(
        orig_height: int, orig_width: int, out_height: int, out_width: int
    ) -> Tuple[int, int, int, int]:
        """Computes the padding required to resize an image to a
        letterbox format.

        @type orig_height: int
        @param orig_height: Original height of the image.
        @type orig_width: int
        @param orig_width: Original width of the image.
        @type out_height: int
        @param out_height: Desired height of the output.
        @type out_width: int
        @param out_width: Desired width of the output.
        @rtype: Tuple[int, int, int, int]
        @return: Padding values for the top, bottom, left and right
            sides of the image.
        """
        ratio = min(out_height / orig_height, out_width / orig_width)
        new_height = int(orig_height * ratio)
        new_width = int(orig_width * ratio)

        pad_top = (out_height - new_height) // 2
        pad_bottom = pad_top

        pad_left = (out_width - new_width) // 2
        pad_right = pad_left
        return pad_top, pad_bottom, pad_left, pad_right

    @override
    def apply(
        self,
        img: np.ndarray,
        pad_top: int,
        pad_bottom: int,
        pad_left: int,
        pad_right: int,
        **_,
    ) -> np.ndarray:
        """Applies the letterbox augmentation to an image.

        @type img: np.ndarray
        @param img: Input image to which resize is applied.
        @type pad_top: int
        @param pad_top: Number of pixels to pad at the top.
        @type pad_bottom: int
        @param pad_bottom: Number of pixels to pad at the bottom.
        @type pad_left: int
        @param pad_left: Number of pixels to pad on the left.
        @type pad_right: int
        @param pad_right: Number of pixels to pad on the right.
        @rtype: np.ndarray
        @return: Image with applied letterbox resize.
        """
        return self._apply_to_image_data(
            img,
            pad_top,
            pad_bottom,
            pad_left,
            pad_right,
            self.interpolation,
            self.image_fill_value,
        )

    @override
    def apply_to_mask(
        self,
        img: np.ndarray,
        pad_top: int,
        pad_bottom: int,
        pad_left: int,
        pad_right: int,
        **_,
    ) -> np.ndarray:
        """Applies letterbox augmentation to the input mask."""
        return self._apply_to_image_data(
            img,
            pad_top,
            pad_bottom,
            pad_left,
            pad_right,
            cv2.INTER_NEAREST,
            self.mask_fill_value,
        )

    @override
    def apply_to_bboxes(
        self,
        bbox: np.ndarray,
        pad_top: int,
        pad_bottom: int,
        pad_left: int,
        pad_right: int,
        **_,
    ) -> np.ndarray:
        """Applies letterbox augmentation to the bounding box."""

        if bbox.size == 0:
            return bbox

        pad_left_norm = pad_left / self.width
        pad_right_norm = pad_right / self.width
        pad_top_norm = pad_top / self.height
        pad_bottom_norm = pad_bottom / self.height

        bbox[:, [0, 2]] *= 1 - pad_left_norm - pad_right_norm
        bbox[:, [0, 2]] += pad_left_norm

        bbox[:, [1, 3]] *= 1 - pad_top_norm - pad_bottom_norm
        bbox[:, [1, 3]] += pad_top_norm

        np.clip(
            bbox[:, [0, 2]],
            pad_left_norm,
            1 - pad_right_norm,
            out=bbox[:, [0, 2]],
        )
        np.clip(
            bbox[:, [1, 3]],
            pad_top_norm,
            1 - pad_bottom_norm,
            out=bbox[:, [1, 3]],
        )

        return bbox

    @override
    def apply_to_keypoints(
        self,
        keypoint: np.ndarray,
        pad_top: int,
        pad_bottom: int,
        pad_left: int,
        pad_right: int,
        cols: int,
        rows: int,
        **_,
    ) -> np.ndarray:
        """Applies letterbox augmentation to the keypoint."""

        if keypoint.size == 0:
            return keypoint

        scale_x = (self.width - pad_left - pad_right) / cols
        scale_y = (self.height - pad_top - pad_bottom) / rows
        keypoint[:, 0] *= scale_x
        keypoint[:, 0] += pad_left

        keypoint[:, 1] *= scale_y
        keypoint[:, 1] += pad_top

        out_of_bounds_x = np.logical_or(
            keypoint[:, 0] < pad_left, keypoint[:, 0] > self.width - pad_right
        )
        out_of_bounds_y = np.logical_or(
            keypoint[:, 1] < pad_top, keypoint[:, 1] > self.height - pad_bottom
        )
        keypoint[out_of_bounds_x | out_of_bounds_y, :2] = -1

        return keypoint

    def _apply_to_image_data(
        self,
        img: np.ndarray,
        pad_top: int,
        pad_bottom: int,
        pad_left: int,
        pad_right: int,
        interpolation: int,
        fill_value: RGB,
    ) -> np.ndarray:
        resized_img = cv2.resize(
            img,
            (
                self.width - pad_left - pad_right,
                self.height - pad_top - pad_bottom,
            ),
            interpolation=interpolation,
        )
        return cv2.copyMakeBorder(
            resized_img,
            pad_top,
            pad_bottom,
            pad_left,
            pad_right,
            cv2.BORDER_CONSTANT,
            value=fill_value,
        ).astype(img.dtype)



from .archive_generator import ArchiveGenerator
from .config import Config
from .model import Model
from .utils import infer_layout, is_nn_archive

__all__ = [
    "ArchiveGenerator",
    "Config",
    "Model",
    "infer_layout",
    "is_nn_archive",
]



import builtins
import platform
import random
import shutil
import sys
import time
from pathlib import Path
from typing import Dict, List

import numpy as np
import pytest
from _pytest.fixtures import SubRequest
from _pytest.main import Session
from rich import print as rich_print

from luxonis_ml.data import BucketStorage, LuxonisDataset
from luxonis_ml.typing import Params
from luxonis_ml.utils.environ import environ

CREATED_DATASETS = []


@pytest.fixture(autouse=True, scope="session")
def setup():
    builtins.print = rich_print

    randint = random.randint(0, 100000)
    environ.LUXONISML_BASE_PATH = (
        Path.cwd() / f"tests/data/luxonisml_base_path/{randint}"
    )
    if environ.LUXONISML_BASE_PATH.exists():  # pragma: no cover
        shutil.rmtree(environ.LUXONISML_BASE_PATH)
    environ.LUXONISML_BASE_PATH.mkdir(parents=True, exist_ok=True)


@pytest.fixture
def randint() -> int:
    return random.randint(0, 100_000)


@pytest.fixture(autouse=True, scope="session")
def fix_seed(worker_id: str):
    np.random.seed(hash(worker_id) % 2**32)
    random.seed(hash(worker_id) % 2**32)


@pytest.fixture(scope="session")
def python_version():
    version = sys.version_info
    return f"{version.major}{version.minor}"


@pytest.fixture(scope="session")
def platform_name():  # pragma: no cover
    os_name = platform.system().lower()
    if "darwin" in os_name:
        return "mac"
    if "linux" in os_name:
        return "lin"
    if "windows" in os_name:
        return "win"
    raise ValueError(f"Unsupported operating system: {os_name}")


@pytest.fixture
def dataset_name(request: SubRequest, randint: int) -> str:
    name = f"{get_caller_name(request)}_{randint}"
    CREATED_DATASETS.append(name)
    return name


def pytest_sessionfinish(session: Session, exitstatus: int) -> None:
    for ds_name in CREATED_DATASETS:
        LuxonisDataset(
            ds_name, bucket_storage=BucketStorage.GCS
        ).delete_dataset(delete_remote=True, delete_local=True)


@pytest.fixture(scope="session")
def height() -> int:
    return 480


@pytest.fixture(scope="session")
def width() -> int:
    return 640


@pytest.fixture
def augmentation_data(
    height: int, width: int, request: pytest.FixtureRequest
) -> Dict[str, List[np.ndarray]]:
    batch_size: int = request.param
    return {
        "image": [
            np.random.rand(height, width, 3) * 255 for _ in range(batch_size)
        ],
        "bboxes": [
            np.array([[0.3 + i * 0.1, 0.3 + i * 0.1, 0.1, 0.1, 0, 0]])
            for i in range(batch_size)
        ],
        "keypoints": [
            np.array([[64.0 + i * 10, 150.0 + i * 10, 0.0, 0.0, 2.0]])
            for i in range(batch_size)
        ],
        "mask": [
            np.random.randint(0, 2, (height, width, 1))
            for _ in range(batch_size)
        ],
    }


@pytest.fixture(scope="session")
def augmentation_config() -> List[Params]:
    return [
        {
            "name": "Mosaic4",
            "params": {"out_width": 416, "out_height": 416, "p": 1.0},
        },
        {"name": "MixUp", "params": {"p": 1.0}},
        {"name": "Defocus", "params": {"p": 1.0}},
        {"name": "Sharpen", "params": {"p": 1.0}},
        {"name": "Flip", "params": {"p": 1.0}},
        {"name": "RandomRotate90", "params": {"p": 1.0}},
    ]


@pytest.fixture(scope="session")
def base_tempdir(worker_id: str):
    path = Path("tests", "data", "tempdir", worker_id)
    shutil.rmtree(path, ignore_errors=True)
    path.mkdir(parents=True, exist_ok=True)
    return path


@pytest.fixture
def tempdir(base_tempdir: Path, randint: int) -> Path:
    t = time.time()
    while True:
        path = base_tempdir / str(randint)
        if not path.exists():
            break
        if time.time() - t > 5:  # pragma: no cover
            raise TimeoutError(
                "Could not create a unique tempdir. Something is wrong."
            )

    path.mkdir(exist_ok=True)

    return path


@pytest.fixture(scope="session")
def storage_url() -> str:
    return "gs://luxonis-test-bucket/luxonis-ml-test-data/"


def pytest_addoption(parser: pytest.Parser):
    parser.addoption(
        "--only-local",
        action="store_true",
        default=False,
        help="Run tests only for local storage",
    )


def pytest_generate_tests(metafunc: pytest.Metafunc):
    if "bucket_storage" in metafunc.fixturenames:
        only_local = metafunc.config.getoption("--only-local")
        storage_options = (
            [BucketStorage.LOCAL]
            if only_local
            else [BucketStorage.LOCAL, BucketStorage.GCS]
        )
        metafunc.parametrize("bucket_storage", storage_options)


def get_caller_name(request: SubRequest) -> str:  # pragma: no cover
    node = request.node
    if isinstance(node, pytest.Function):
        return node.function.__name__
    return node.name



"""MNIST Dataset Loader with Custom Preprocessing.

This module provides utility functions to load the MNIST dataset with specific
transformations applied. The dataset is transformed to be compatible with models
that expect 3-channel RGB images, such as models pre-trained on the ImageNet dataset.

The main transformations applied are:
    1. Convert single-channel grayscale images to 3-channel grayscale.
    2. Convert 3-channel grayscale images to RGB format (though the image will still look grayscale visually).
    3. Resize images to the size of (224, 224) to be compatible with models like ResNet-50.
    4. Convert images to PyTorch tensors.
    5. Normalize the tensor values using the mean and standard deviation of ImageNet.

Functions:
    - `mnist_transformations()`: Returns the composed transformations.
    - `load_mnist_data(save_path='./mnist', num_samples=640, batch_size=64)`: Loads MNIST data with the defined transformations
        and returns a DataLoader. It provides options to specify the number of samples and batch size.

Example usage:
    ```
    data_loader = load_mnist_data('./mnist_data', num_samples=1000, batch_size=32)
    for batch in data_loader:
        images, labels = batch
        ...  # Your processing here
    ```

Note: This loader is particularly useful when you want to use MNIST data with models that were
pre-trained on datasets like ImageNet and expect 3-channel RGB input.
"""

import torch
import torchvision
from torchvision import transforms


def mnist_transformations() -> transforms.Compose:
    """Returns composed transformations for the MNIST dataset.

    Transforms the images from 1 channel grayscale to 3 channels RGB and
    resizes them.
    """
    return transforms.Compose(
        [
            transforms.Grayscale(num_output_channels=3),
            transforms.Lambda(lambda x: x.convert("RGB")),
            transforms.Resize((224, 224)),
            transforms.ToTensor(),
            transforms.Normalize(
                mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
            ),
        ]
    )


def load_mnist_data(
    save_path: str = "./mnist", num_samples: int = 640, batch_size: int = 64
) -> torch.utils.data.DataLoader:
    """Loads the MNIST dataset with the specified preprocessing.

    Parameters:
    - save_path (str): Directory to save/load the MNIST data.
    - num_samples (int): Number of samples to load from the dataset.
                         Set as -1 to load the entire dataset.
    - batch_size (int): Batch size for the DataLoader.

    Returns:
    - data_loader (DataLoader): DataLoader for the MNIST dataset.
    """
    transform = mnist_transformations()

    # Load the MNIST dataset
    dataset = torchvision.datasets.MNIST(
        root=save_path, train=True, transform=transform, download=True
    )

    # If num_samples is set to -1, use the entire dataset
    num_samples = (
        min(num_samples, len(dataset)) if num_samples != -1 else len(dataset)
    )

    # Create a subset of the dataset using Subset class
    subset = torch.utils.data.Subset(dataset, torch.arange(num_samples))

    # Create a data loader to load the dataset in batches
    data_loader = torch.utils.data.DataLoader(
        subset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=4,
        pin_memory=True,
    )

    return data_loader



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List

import albumentations as A
import numpy as np
from typing_extensions import override


class BatchTransform(ABC, A.DualTransform):
    def __init__(self, batch_size: int, **kwargs):
        """Batch transformation that combines multiple images and
        associated labels into one.

        @param batch_size: Batch size needed for augmentation to work
        @type batch_size: int
        """
        super().__init__(**kwargs)

        self.batch_size = batch_size

    @property
    @override
    def targets(self) -> Dict[str, Callable]:
        targets = super().targets
        targets.update(
            {
                "instance_mask": self.apply_to_instance_mask,
                "array": self.apply_to_array,
                "classification": self.apply_to_classification,
                "metadata": self.apply_to_metadata,
            }
        )
        return targets

    @abstractmethod
    def apply(self, image_batch: List[np.ndarray], **kwargs) -> np.ndarray: ...

    @abstractmethod
    def apply_to_mask(
        self, mask_batch: List[np.ndarray], **kwargs
    ) -> np.ndarray: ...

    @abstractmethod
    def apply_to_bboxes(
        self, bboxes_batch: List[np.ndarray], **kwargs
    ) -> np.ndarray: ...

    @abstractmethod
    def apply_to_keypoints(
        self, keypoints_batch: List[np.ndarray], **kwargs
    ) -> np.ndarray: ...

    @abstractmethod
    def apply_to_instance_mask(
        self, masks_batch: List[np.ndarray], **kwargs
    ) -> np.ndarray: ...

    def apply_to_array(self, array_batch: List[np.ndarray], **_) -> np.ndarray:
        return np.concatenate([arr for arr in array_batch if arr.size > 0])

    def apply_to_classification(
        self, classification_batch: List[np.ndarray], **_
    ) -> np.ndarray:
        for i in range(len(classification_batch)):
            if classification_batch[i].size == 0:
                classification_batch[i] = np.zeros(1)
        return np.clip(sum(classification_batch), 0, 1)

    def apply_to_metadata(
        self, metadata_batch: List[np.ndarray], **_
    ) -> np.ndarray:
        if all(arr.size == 0 for arr in metadata_batch):
            return np.array([])
        return np.concatenate([arr for arr in metadata_batch if arr.size > 0])

    @override
    def update_params(self, params: Dict[str, Any], **_) -> Dict[str, Any]:
        return params

    @override
    def update_params_shape(
        self, params: Dict[str, Any], data: Dict[str, Any]
    ) -> Dict[str, Any]:
        shape = data["image"][0].shape
        params["shape"] = shape
        params.update({"cols": shape[1], "rows": shape[0]})
        return params



from typing import Final

import numpy as np
import pytest
from albumentations.core.bbox_utils import normalize_bboxes
from pytest_subtests.plugin import SubTests

from luxonis_ml.data.augmentations.custom.mosaic import (
    Mosaic4,
    apply_mosaic4_to_bboxes,
    apply_mosaic4_to_images,
    apply_mosaic4_to_keypoints,
)

WIDTH: Final[int] = 640
HEIGHT: Final[int] = 480


def test_mosaic4_helpers(subtests: SubTests):
    with subtests.test("image"):
        img = (np.random.rand(HEIGHT, WIDTH, 3) * 255).astype(np.uint8)
        mosaic = apply_mosaic4_to_images(
            [img, img, img, img], HEIGHT, WIDTH, 0, 0, 0
        )
        assert mosaic.shape == (HEIGHT, WIDTH, 3)

    with subtests.test("bboxes"):
        bbox = np.array([0, 0, WIDTH, HEIGHT])[None, ...]
        expected_bbox = [
            [0, 0, 1, 1],
            [1, 0, 2, 1],
            [0, 1, 1, 2],
            [1, 1, 2, 2],
        ]
        for i in range(4):
            normalized_bbox = normalize_bboxes(bbox, (HEIGHT, WIDTH))
            mosaic_bbox = apply_mosaic4_to_bboxes(
                normalized_bbox,
                HEIGHT // 2,
                WIDTH // 2,
                i,
                HEIGHT,
                WIDTH,
                0,
                0,
            )[0].tolist()
            assert pytest.approx(mosaic_bbox, abs=0.01) == expected_bbox[i]

    with subtests.test("keypoints"):
        expected_kpts = [
            [WIDTH, HEIGHT, 0, 2],
            [3 * WIDTH, HEIGHT, 0, 0],
            [WIDTH, 3 * HEIGHT, 0, 0],
            [3 * WIDTH, 3 * HEIGHT, 0, 0],
        ]
        for i, (w, h) in enumerate(
            [
                (WIDTH // 2, HEIGHT // 2),
                (WIDTH, HEIGHT // 2),
                (WIDTH // 2, HEIGHT),
                (WIDTH, HEIGHT),
            ]
        ):
            mosaic_keypoint = apply_mosaic4_to_keypoints(
                np.array([w, h, 0, 2])[None, ...],
                HEIGHT // 2,
                WIDTH // 2,
                i,
                HEIGHT,
                WIDTH,
                0,
                0,
            )[0].tolist()
            assert pytest.approx(mosaic_keypoint, abs=0.1) == expected_kpts[i]


def test_mosaic4():
    img = (np.random.rand(HEIGHT, WIDTH, 3) * 255).astype(np.uint8)
    mosaic4 = Mosaic4(out_height=HEIGHT, out_width=WIDTH, p=1.0)
    m = mosaic4(image=[img, img, img, img])
    assert m["image"].shape == (HEIGHT, WIDTH, 3)


def test_invalid():
    with pytest.raises(ValueError, match="`out_height` must be larger"):
        Mosaic4(out_height=0, out_width=WIDTH)

    with pytest.raises(ValueError, match="`out_width` must be larger"):
        Mosaic4(out_height=HEIGHT, out_width=0)



"""Mismatch Detection in Labelled Data.

This module provides functionalities to detect mismatches or potential mislabelling
in a dataset based on various strategies. This is crucial in supervised machine
learning tasks where the quality of labels significantly affects model performance.

Methods implemented
===================
    - Centroids: This method identifies mismatches by comparing the distance of data points
                 to the centroid of their own class against the distances to centroids of
                 other classes.
    - KNN (k-Nearest Neighbors): This approach leverages the idea that if the majority of data
                                 is correctly labelled, then mislabelled data will be corrected
                                 by its nearest neighbors.
    - [Note: DBSCAN was considered but not implemented due to underperformance.]

Usage
=====
    To use this module, import the desired methods and provide the embeddings and labels:

        >>> from mismatch_detection import find_mismatches_centroids, find_mismatches_knn
        >>> # Detect mismatches using centroids
        >>> mismatches, new_labels = find_mismatches_centroids(X_train, y_train)
        >>> # Detect mismatches using KNN
        >>> mismatches, new_labels = find_mismatches_knn(X_train, y_train)
"""

from typing import Tuple

import numpy as np
from scipy.spatial import distance
from sklearn.neighbors import KNeighborsClassifier


def find_mismatches_centroids(
    X: np.array, y: np.array
) -> Tuple[np.array, np.array]:
    """Find mismatches in the dataset. A mismatch is defined as a sample
    that is closer to another centroid than to its own centroid.

    @type X: np.array
    @param X: The embeddings to use.
    @type y: np.array
    @param y: The targets to use.
    @rtype: Tuple[np.array, np.array]
    @return: The indices of the mismatches and the new labels.
    """
    unique_labels = np.unique(y)
    # Create a mapping from string labels to integer indices
    label_to_index = {
        label: index for index, label in enumerate(unique_labels)
    }

    # calculate centroids of each class
    centroids = []
    for label in unique_labels:
        # Get the indices of the samples in the train set that belong to the specified class
        indices = np.where(y == label)[0]

        # Get the embeddings of the samples in the specified class
        embeddings_class = X[indices]

        # Calculate the centroid of the specified class
        centroid = np.mean(embeddings_class, axis=0)
        centroids.append(centroid)

    # append the centroids to the embeddings
    centroids = np.array(centroids)

    # make the distance matrix
    dist_matrix = distance.cdist(X, centroids, metric="cosine")

    # find mismatches if the distance to other centroids is 1.5 times larger than the distance to their own centroid
    mismatches = []
    predicted_labels = []
    for i in range(len(X)):
        # find the distance to the closest centroid
        closest_centroid = np.argmin(dist_matrix[i, :])
        closest_distance = dist_matrix[i, closest_centroid]

        # find the distance to their own centroid
        own_centroid = dist_matrix[i, label_to_index[y[i]]]

        # if the distance to the closest centroid is 1.5 times larger than the distance to their own centroid, then it is a mismatch
        if closest_distance * 1.5 < own_centroid:
            mismatches.append(i)
            predicted_labels.append(unique_labels[closest_centroid])
        # else:
        #     predicted_labels.append(y[i])

    mismatches = np.array(mismatches)
    predicted_labels = np.array(predicted_labels)
    new_labels = predicted_labels

    return mismatches, new_labels


def find_mismatches_knn(
    X: np.array, y: np.array, n_neighbors: int = 5
) -> Tuple[np.array, np.array]:
    """
    Find mismatches in the dataset.
    Single Algorithm Filter (see Figure 1 in Brodley, Carla E., and Mark A. Friedl. "Identifying mislabeled training data.").
    Idea: if the vast majority of the data is correctly labeled and you do knn prediction, the minority of mislabeled data will be engulfed (corrected) by the correct neighbors.

    @type X: np.array
    @param X: The embeddings to use.

    @type y: np.array
    @param y: The targets to use.

    @type n_neighbors: int
    @param n_neighbors: The number of neighbors to use for KNN. Default is 5.

    @rtype: Tuple[np.array, np.array]
    @return: The indices of the mismatches and the new labels.
    """

    # Step 1: Fit KNN on the train set with corrupted labels
    knn = KNeighborsClassifier(n_neighbors=n_neighbors)
    knn.fit(X, y)

    # Step 2: Predict labels on the train set using the trained KNN model
    y_pred = knn.predict(X)

    # Step 3: Find the mismatched labels
    mismatches = np.where(y_pred != y)[0]

    new_labels = y_pred

    return mismatches, new_labels



from copy import deepcopy
from typing import Dict, Iterator, List, Mapping, Set, Tuple, TypeVar

T = TypeVar("T")


def traverse_graph(
    graph: Dict[str, List[str]], nodes: Mapping[str, T]
) -> Iterator[Tuple[str, T, List[str], List[str]]]:
    """Traverses the graph in topological order, starting from the nodes
    with no predecessors.

    Example:

        >>> graph = {"a": ["b"], "b": []}
        >>> nodes = {"a": 1, "b": 2}
        >>> for name, value, preds, rem in traverse_graph(graph, nodes):
        ...     print(name, value, preds, rem)
        b 2 [] ['a']
        a 1 ['b'] []

    @type graph: dict[str, list[str]]
    @param graph: Graph in a format of a dictionary of predecessors.
        Keys are node names, values are node predecessors (list of node
        names).
    @type nodes: dict[str, T]
    @param nodes: Dictionary mapping node names to values.
    @rtype: Iterator[tuple[str, T, list[str], list[str]]]
    @return: Iterator of tuples containing node name, node value, node
        predecessors, and remaining unprocessed nodes.
    @raises RuntimeError: If the graph is malformed.
    """
    # sort the set for consistent behavior
    unprocessed_nodes = sorted(set(nodes.keys()))
    processed: set[str] = set()

    graph = deepcopy(graph)
    while unprocessed_nodes:
        unprocessed_nodes_copy = unprocessed_nodes.copy()
        for node_name in unprocessed_nodes_copy:
            node_dependencies = graph[node_name]
            if not node_dependencies or all(
                dependency in processed for dependency in node_dependencies
            ):
                unprocessed_nodes.remove(node_name)
                yield (
                    node_name,
                    nodes[node_name],
                    node_dependencies,
                    unprocessed_nodes.copy(),
                )
                processed.add(node_name)

        if unprocessed_nodes_copy == unprocessed_nodes:
            raise RuntimeError(
                "Malformed graph. "
                "Please check that all nodes are connected in a directed acyclic graph."
            )


def is_acyclic(graph: Dict[str, List[str]]) -> bool:
    """Tests if graph is acyclic.

    @type graph: dict[str, list[str]]
    @param graph: Graph in a format of a dictionary of predecessors.
    @rtype: bool
    @return: True if graph is acyclic, False otherwise.
    """
    graph = graph.copy()

    def dfs(node: str, visited: Set[str], recursion_stack: Set[str]) -> bool:
        visited.add(node)
        recursion_stack.add(node)

        for predecessor in graph.get(node, []):
            if predecessor in recursion_stack:
                return True
            if predecessor not in visited and dfs(
                predecessor, visited, recursion_stack
            ):
                return True

        recursion_stack.remove(node)
        return False

    visited: set[str] = set()
    recursion_stack: set[str] = set()

    for node in graph:
        if node not in visited and dfs(node, visited, recursion_stack):
            return False

    return True



import random
import shutil
from pathlib import Path
from typing import Iterator, List, Optional, Set, Tuple

import cv2
import matplotlib.pyplot as plt
import numpy as np
import rich.box
import typer
from rich import print
from rich.console import Console, RenderableType, group
from rich.panel import Panel
from rich.prompt import Confirm
from rich.rule import Rule
from rich.table import Table
from typing_extensions import Annotated

from luxonis_ml.data import (
    LuxonisDataset,
    LuxonisLoader,
    LuxonisParser,
    UpdateMode,
)
from luxonis_ml.data.utils.enums import BucketStorage
from luxonis_ml.data.utils.plot_utils import (
    plot_class_distribution,
    plot_heatmap,
)
from luxonis_ml.data.utils.visualizations import visualize
from luxonis_ml.enums import DatasetType

app = typer.Typer()


def complete_dataset_name(incomplete: str):
    datasets = LuxonisDataset.list_datasets()
    return [d for d in datasets if d.startswith(incomplete)]


DatasetNameArgument = Annotated[
    str,
    typer.Argument(
        ...,
        help="Name of the dataset.",
        autocompletion=complete_dataset_name,
        show_default=False,
    ),
]

bucket_option = typer.Option(
    "local",
    "--bucket-storage",
    "-b",
    help="Storage type for the dataset.",
)


def check_exists(name: str, bucket_storage: BucketStorage):
    if not LuxonisDataset.exists(name, bucket_storage=bucket_storage):
        print(f"[red]Dataset [magenta]'{name}'[red] does not exist.")
        raise typer.Exit


def get_dataset_info(dataset: LuxonisDataset) -> Tuple[Set[str], List[str]]:
    all_classes = {
        c for classes in dataset.get_classes().values() for c in classes
    }
    return all_classes, dataset.get_task_names()


def print_info(dataset: LuxonisDataset) -> None:
    classes = dataset.get_classes()
    has_named_classes = any(k for k in classes if k)
    class_table = Table(
        title="Classes", box=rich.box.ROUNDED, row_styles=["yellow", "cyan"]
    )
    if has_named_classes:
        class_table.add_column(
            "Task Name", header_style="magenta i", max_width=30
        )
    class_table.add_column(
        "Class Names", header_style="magenta i", max_width=50
    )

    for task_name, c in classes.items():
        if has_named_classes:
            class_table.add_row(task_name, ", ".join(c))
        else:
            class_table.add_row(", ".join(c))

    tasks = dataset.get_tasks()
    has_named_tasks = any(k for k in tasks if k)

    task_table = Table(
        title="Tasks", box=rich.box.ROUNDED, row_styles=["yellow", "cyan"]
    )
    if has_named_tasks:
        task_table.add_column(
            "Task Name", header_style="magenta i", max_width=30
        )
    task_table.add_column("Task Types", header_style="magenta i", max_width=50)
    for task_name, task_types in tasks.items():
        task_types.sort()
        if has_named_tasks:
            task_table.add_row(task_name, ", ".join(task_types))
        else:
            task_table.add_row(", ".join(task_types))

    splits = dataset.get_splits()

    @group()
    def get_sizes_panel() -> Iterator[RenderableType]:
        if splits is not None:
            total_files = len(dataset)
            for split, files in splits.items():
                split_size = len(files)
                percentage = (
                    (split_size / total_files * 100) if total_files > 0 else 0
                )
                yield f"[magenta b]{split}: [not b cyan]{split_size:,} [dim]({percentage:.1f}%)[/dim]"
        else:
            yield "[red]No splits found"
        yield Rule()
        yield f"[magenta b]Total: [not b cyan]{len(dataset)}"

    @group()
    def get_panels() -> Iterator[RenderableType]:
        yield f"[magenta b]Name: [not b cyan]{dataset.identifier}"
        yield f"[magenta b]Version: [not b cyan]{dataset.version}"
        yield ""
        yield Panel.fit(get_sizes_panel(), title="Split Sizes")
        yield class_table
        yield task_table

    print(Panel.fit(get_panels(), title="Dataset Info"))


@app.command()
def info(
    name: DatasetNameArgument,
    bucket_storage: BucketStorage = bucket_option,
):
    """Prints information about a dataset."""
    check_exists(name, bucket_storage)
    print_info(LuxonisDataset(name, bucket_storage=bucket_storage))


@app.command()
def delete(
    name: DatasetNameArgument,
    bucket_storage: BucketStorage = bucket_option,
    local: bool = typer.Option(
        False,
        "--local/--no-local",
        help="Delete the dataset from local storage.",
    ),
    remote: bool = typer.Option(
        False,
        "--remote/--no-remote",
        help="Delete the dataset from remote storage.",
    ),
):
    """Deletes a dataset from local storage or remote storage (or both),
    based on which options are passed."""
    check_exists(name, bucket_storage)

    if bucket_storage is BucketStorage.LOCAL and remote:
        print(
            "[yellow]Warning: You specified remote deletion, but the bucket is local. "
            "Remote deletion will not be performed.[/yellow]"
        )
        remote = False

    where = " and ".join(
        filter(None, ["local" if local else "", "remote" if remote else ""])
    )

    if not where:
        print(
            "[yellow]No deletion target specified (local or remote). Nothing to delete.[/yellow]"
        )
        raise typer.Exit

    if not Confirm.ask(
        f"Delete dataset '{name}' with specified bucket '{bucket_storage}' from {where} storage?"
    ):
        raise typer.Exit

    dataset = LuxonisDataset(name, bucket_storage=bucket_storage)
    dataset.delete_dataset(
        delete_local=local,
        delete_remote=remote,
    )

    print(
        f"Dataset '{name}' deleted from: "
        f"{'local ' if local else ''}"
        f"{'remote ' if remote else ''}"
        f"storage."
    )


@app.command()
def ls(
    full: bool = typer.Option(
        False, "--full", "-f", help="Show full information."
    ),
    bucket_storage: BucketStorage = bucket_option,
):
    """Lists all datasets."""
    datasets = LuxonisDataset.list_datasets(bucket_storage=bucket_storage)
    table = Table(
        title="Datasets" + (" - Full Table" if full else ""),
        box=rich.box.ROUNDED,
        row_styles=["yellow", "cyan"],
    )
    table.add_column("Name", header_style="magenta i")
    table.add_column("Size", header_style="magenta i")
    if full:
        table.add_column("Classes", header_style="magenta i")
        table.add_column("Tasks", header_style="magenta i")
    for name in datasets:
        dataset = LuxonisDataset(name, bucket_storage=bucket_storage)
        rows = [name]
        try:
            size = len(dataset)
        except Exception:
            size = -1
        rows.append(str(size))
        if full:
            classes, tasks = get_dataset_info(dataset)
            rows.extend(
                [
                    ", ".join(classes) if classes else "[red]<empty>[no red]",
                    ", ".join(tasks) if tasks else "[red]<empty>[no red]",
                ]
            )
        table.add_row(*rows)
    console = Console()
    console.print(table)


@app.command()
def inspect(
    name: DatasetNameArgument,
    view: Annotated[
        Optional[List[str]],
        typer.Option(
            ...,
            "--view",
            "-v",
            help="Which splits of the dataset to inspect.",
            case_sensitive=False,
            show_default=False,
        ),
    ] = None,
    aug_config: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--aug-config",
            "-a",
            help="Path to a config defining augmentations. "
            "This can be either a json or a yaml file.",
            metavar="PATH",
            show_default=False,
        ),
    ] = None,
    size_multiplier: Annotated[
        float,
        typer.Option(
            ...,
            "--size-multiplier",
            "-s",
            help=(
                "Multiplier for the image size. "
                "By default the images are shown in their original size."
            ),
            show_default=False,
        ),
    ] = 1.0,
    ignore_aspect_ratio: Annotated[
        bool,
        typer.Option(
            ...,
            "--ignore-aspect-ratio",
            "-i",
            help="Don't keep the aspect ratio when resizing images.",
        ),
    ] = False,
    deterministic: Annotated[
        bool,
        typer.Option(
            ...,
            "--deterministic",
            "-d",
            help="Deterministic mode. Useful for debugging.",
        ),
    ] = False,
    force_update: Annotated[
        bool,
        typer.Option(
            ...,
            "--force-update",
            "-f",
            help="Force synchronization of the dataset with the "
            "remote storage. Only relevant for remote datasets.",
        ),
    ] = False,
    blend_all: Annotated[
        bool,
        typer.Option(
            ...,
            "--blend-all",
            "-b",
            help="Whether to draw labels belonging "
            "to different tasks on the same image. "
            "Doesn't apply to semantic segmentations.",
        ),
    ] = False,
    per_instance: Annotated[
        bool,
        typer.Option(
            ...,
            "--per-instance",
            "-pi",
            help="Show each label instance in a separate window.",
        ),
    ] = False,
    bucket_storage: BucketStorage = bucket_option,
):
    """Inspects images and annotations in a dataset."""

    if deterministic:
        np.random.seed(42)
        random.seed(42)

    view = view or ["train"]
    dataset = LuxonisDataset(name, bucket_storage=bucket_storage)
    loader = LuxonisLoader(
        dataset,
        view=view,
        update_mode="all" if force_update else "missing",
    )

    if aug_config is not None:
        h, w, _ = loader[0][0].shape
        loader.augmentations = loader._init_augmentations(
            "albumentations", aug_config, h, w, not ignore_aspect_ratio
        )

    if len(dataset) == 0:
        raise ValueError(f"Dataset '{name}' is empty.")

    classes = dataset.get_classes()
    for image, labels in loader:
        image = image.astype(np.uint8)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        h, w, _ = image.shape
        new_h, new_w = int(h * size_multiplier), int(w * size_multiplier)
        image = cv2.resize(image, (new_w, new_h))
        instance_keys = [
            "/boundingbox",
            "/keypoints",
            "/instance_segmentation",
        ]
        matched_instance_keys = [
            k for k in labels if any(k.endswith(ik) for ik in instance_keys)
        ]
        if per_instance and matched_instance_keys:
            extra_keys = [k for k in labels if k not in matched_instance_keys]
            if extra_keys:
                print(
                    f"[yellow]Warning: Ignoring non-instance keys in labels: {extra_keys}[/yellow]"
                )
            n_instances = len(labels[matched_instance_keys[0]])
            for i in range(n_instances):
                instance_labels = {
                    k: np.expand_dims(v[i], axis=0)
                    for k, v in labels.items()
                    if k in matched_instance_keys and len(v) > i
                }
                instance_image = visualize(
                    image.copy(), instance_labels, classes, blend_all=blend_all
                )
                cv2.imshow("image", instance_image)
                if cv2.waitKey() == ord("q"):
                    break
        else:
            if per_instance:
                print(
                    "[yellow]Warning: Per-instance mode is not supported for this dataset. "
                    "Showing all labels in one window.[/yellow]"
                )
            image = visualize(image, labels, classes, blend_all=blend_all)
            cv2.imshow("image", image)
            if cv2.waitKey() == ord("q"):
                break


@app.command()
def export(
    dataset_name: DatasetNameArgument,
    save_dir: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--save-dir",
            "-s",
            help="Directory where the dataset should be saved. "
            "If not provided, the dataset will be saved in the "
            "current working directory under the name of the dataset.",
            show_default=False,
        ),
    ] = None,
    dataset_type: Annotated[
        DatasetType,
        typer.Option(
            ...,
            "--type",
            "-t",
            help="Format of the exported dataset",
            show_default=False,
        ),
    ] = "native",  # type: ignore
    delete_existing: Annotated[
        bool,
        typer.Option(
            ...,
            "--delete",
            "-d",
            help="Delete an existing `save_dir` before exporting.",
        ),
    ] = False,
    task_name_to_keep: Annotated[
        Optional[str],
        typer.Option(
            "--task-name",
            "-tn",
            help=(
                "Name of the single task to export. "
                "Required when the dataset contains multiple tasks; "
                "ignored if the dataset has exactly one task."
            ),
            show_default=False,
        ),
    ] = None,
    max_partition_size_gb: Annotated[
        Optional[float],
        typer.Option(
            ...,
            "--max-partition-size-gb",
            "-m",
            help=(
                "Maximum size of each partition in GB. If the dataset"
                " exceeds this size, it will be split into multiple partitions named {dataset_name}_part{partition_number}."
                " Default is None, meaning the dataset will be exported as a single partition named {dataset_name}."
            ),
            show_default=False,
        ),
    ] = None,
    no_zip: Annotated[
        bool,
        typer.Option(
            "--no-zip",
            help="Skip zipping the exported dataset. By default, the dataset (or each partition) will be zipped.",
        ),
    ] = False,
    bucket_storage: BucketStorage = bucket_option,
):
    save_dir = save_dir or dataset_name
    if delete_existing and Path(save_dir).exists():
        shutil.rmtree(save_dir)
    dataset = LuxonisDataset(dataset_name, bucket_storage=bucket_storage)
    dataset.export(
        save_dir,
        dataset_type,
        task_name_to_keep,
        max_partition_size_gb,
        not no_zip,
    )


@app.command()
def parse(
    dataset_dir: Annotated[
        str, typer.Argument(..., help="Path or URL to the dataset.")
    ],
    name: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--name",
            "-n",
            autocompletion=complete_dataset_name,
            help="Name of the dataset.",
            show_default=False,
        ),
    ] = None,
    dataset_type: Annotated[
        Optional[DatasetType],
        typer.Option(
            ...,
            "--type",
            "-t",
            help="Type of the dataset. If not provided, "
            "the parser will try to recognize it automatically.",
            show_default=False,
        ),
    ] = None,
    delete_local: Annotated[
        bool,
        typer.Option(
            ...,
            "--delete",
            "-d",
            help="If an existing local dataset with the same name should "
            "be deleted before parsing.",
        ),
    ] = False,
    save_dir: Annotated[
        Optional[Path],
        typer.Option(
            ...,
            "--save-dir",
            "-s",
            help="If a remote URL is provided in 'dataset_dir', "
            "the dataset will be downloaded to this directory. "
            "Otherwise, the dataset will be downloaded to the "
            "current working directory.",
            show_default=False,
        ),
    ] = None,
    task_name: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--task-name",
            "-tn",
            help="Name of the task that should be used with this dataset. "
            "If not provided, the name of the dataset format will be used.",
            show_default=False,
        ),
    ] = None,
):
    """Parses a directory with data and creates Luxonis dataset."""
    parser = LuxonisParser(
        dataset_dir,
        dataset_name=name,
        dataset_type=dataset_type,
        delete_local=delete_local,
        save_dir=save_dir,
        task_name=task_name,
    )
    dataset = parser.parse()

    print()
    print(Rule())
    print()
    print_info(dataset)


@app.command()
def health(
    name: DatasetNameArgument,
    view: Optional[str] = typer.Option(
        None,
        "--view",
        "-v",
        help="Which splits of the dataset to inspect. If not provided, all dataset will be used.",
        show_default=False,
    ),
    sample_size: Optional[int] = typer.Option(
        None,
        "--sample-size",
        "-n",
        help="Number of annotation rows to sample from the dataset. Note that each task type annotation is in a separate row.",
        show_default=False,
    ),
    save_dir: Optional[str] = typer.Option(
        None,
        "--save-dir",
        "-s",
        help="Directory where the plots should be saved. "
        "If not provided, the plots will be displayed.",
        show_default=False,
    ),
    bucket_storage: BucketStorage = bucket_option,
):
    """Plots class distributions and heatmaps for every task type and
    corresponding task name in the dataset.

    Also checks for files with missing annotations, files that share the
    same UUIDs, and files with duplicate annotations.
    """
    check_exists(name, bucket_storage)
    dataset = LuxonisDataset(name, bucket_storage=bucket_storage)
    stats = dataset.get_statistics(sample_size=sample_size, view=view)
    console = Console()

    missing_annotations = stats["missing_annotations"]
    duplicate_uuids = stats["duplicates"]["duplicate_uuids"]
    duplicate_annotations = stats.get("duplicates", {}).get(
        "duplicate_annotations", []
    )

    if duplicate_uuids:
        duuid_table = Table(
            title="Duplicate UUIDs", box=rich.box.ROUNDED, row_styles=["red"]
        )
        duuid_table.add_column("UUID", style="magenta")
        duuid_table.add_column("Files", style="cyan")

        for item in duplicate_uuids:
            duuid_table.add_row(item["uuid"], ", ".join(item["files"]))

        console.print(duuid_table)

    if duplicate_annotations:
        dann_table = Table(
            title="Duplicate Annotations",
            box=rich.box.ROUNDED,
            row_styles=["red"],
        )
        dann_table.add_column("File Name", style="cyan")
        dann_table.add_column("Task Name", style="magenta")
        dann_table.add_column("Task Type", style="magenta")
        dann_table.add_column("Annotation", style="yellow")
        dann_table.add_column("Count", style="green")

        for item in duplicate_annotations:
            dann_table.add_row(
                item["file_name"],
                item["task_name"],
                item["task_type"],
                str(item["annotation"]),
                str(item["count"]),
            )

        console.print(dann_table)

    if missing_annotations:
        missing_table = Table(
            title="Files With Missing Annotations",
            box=rich.box.ROUNDED,
            row_styles=["yellow"],
        )
        missing_table.add_column("File Name", style="cyan")

        for file in missing_annotations:
            missing_table.add_row(file)

        console.print(missing_table)

    summary_table = Table(
        title="Dataset Health Summary",
        box=rich.box.ROUNDED,
        show_header=False,
        row_styles=["cyan", "yellow", "green"],
    )
    summary_table.add_column("Metric")
    summary_table.add_column("Count")
    summary_table.add_row(
        "Files with missing annotations", str(len(missing_annotations))
    )
    summary_table.add_row(
        "Files with duplicate UUIDs", str(len(duplicate_uuids))
    )
    summary_table.add_row(
        "Files with duplicate annotations", str(len(duplicate_annotations))
    )

    console.print(summary_table)

    all_task_names = sorted(
        set(stats["class_distributions"].keys())
        | set(stats["heatmaps"].keys())
    )
    if not all_task_names:
        console.print("[info]No plots to display.[/info]")
        return

    for task_name in all_task_names:
        class_dist_by_type = stats["class_distributions"].get(task_name, {})
        heatmaps_by_type = stats["heatmaps"].get(task_name, {})
        all_task_types = sorted(
            set(class_dist_by_type.keys()) | set(heatmaps_by_type.keys())
        )

        if not all_task_types:
            console.print(f"[info]No plots for task name: {task_name}[/info]")
            continue

        nrows = len(all_task_types)
        square_size = 4
        fig, axs = plt.subplots(
            nrows, 2, figsize=(square_size * 2, square_size * nrows)
        )
        if task_name != "":
            fig.suptitle(f"Task Name: {task_name}", fontsize=14)

        if nrows == 1:
            axs = [axs]

        for i, task_type in enumerate(all_task_types):
            plot_class_distribution(
                axs[i][0], task_type, class_dist_by_type.get(task_type, [])
            )
            plot_heatmap(
                axs[i][1],
                fig,  # type: ignore
                task_type,
                heatmaps_by_type.get(task_type),  # type: ignore
            )

        plt.tight_layout()
        plt.subplots_adjust(top=0.9, hspace=0.5)

        if save_dir:
            fig.savefig(f"{save_dir}/dataset_health_{task_name}.png", dpi=150)  # type: ignore
            plt.close(fig)
        else:
            plt.show(block=False)
            if plt.waitforbuttonpress():
                plt.close(fig)


@app.command()
def push(
    name: DatasetNameArgument,
    force_update: Annotated[
        bool,
        typer.Option(
            ...,
            "--force-update",
            "-f",
            help="Force pushing all media files, even if they already exist in the cloud. Annotations and metadata will always be pushed.",
        ),
    ] = False,
    target_bucket_storage: BucketStorage = bucket_option,
):
    """Push a local dataset to cloud storage."""
    check_exists(name, BucketStorage.LOCAL)
    dataset = LuxonisDataset(name, bucket_storage=BucketStorage.LOCAL)

    if target_bucket_storage == BucketStorage.LOCAL:
        print(
            "[red]Cannot push to LOCAL storage. Please specify a cloud target."
        )
        raise typer.Exit(1)

    if LuxonisDataset.exists(
        name, bucket_storage=target_bucket_storage
    ) and not Confirm.ask(
        f"Dataset '{name}' already exists in {target_bucket_storage} bucket. If you are unsure about the dataset, please delete it from the cloud storage and try again. Do you want to overwrite it?"
    ):
        raise typer.Exit

    print(
        f"Pushing dataset '{name}' to {target_bucket_storage.value} storage..."
    )

    update_mode = UpdateMode.ALL if force_update else UpdateMode.MISSING
    dataset.push_to_cloud(
        bucket_storage=target_bucket_storage, update_mode=update_mode
    )

    print(
        f"[green]Dataset '{name}' successfully pushed to {target_bucket_storage.value}."
    )


@app.command()
def pull(
    name: DatasetNameArgument,
    force_update: Annotated[
        bool,
        typer.Option(
            ...,
            "--force-update",
            "-f",
            help="Force pulling all media files, even if they already exist locally.",
        ),
    ] = False,
    bucket_storage: BucketStorage = bucket_option,
):
    """Pull a remote dataset to local storage."""
    if bucket_storage == BucketStorage.LOCAL:
        print(
            "[red]Cannot pull from LOCAL storage. Please specify a cloud source."
        )
        raise typer.Exit(1)

    if not LuxonisDataset.exists(name, bucket_storage=bucket_storage):
        print(
            f"[red]Dataset '{name}' does not exist in {bucket_storage.value} storage."
        )
        raise typer.Exit

    print(f"Pulling dataset '{name}' from {bucket_storage.value} storage...")

    update_mode = UpdateMode.ALL if force_update else UpdateMode.MISSING
    dataset = LuxonisDataset(name, bucket_storage=bucket_storage)
    dataset.pull_from_cloud(update_mode=update_mode)

    print(
        f"[green]Dataset '{name}' successfully pulled from {bucket_storage.value}."
    )


@app.command()
def clone(
    name: DatasetNameArgument,
    new_name: Annotated[
        str,
        typer.Argument(
            ..., help="Name of the new dataset.", show_default=False
        ),
    ],
    push_to_cloud: Annotated[
        bool,
        typer.Option(
            "--push/--no-push",
            help="Whether to upload the newly cloned remote dataset back to cloud storage.",
            show_default=True,
        ),
    ] = True,
    bucket_storage: BucketStorage = bucket_option,
):
    """Clone an existing dataset with a new name.

    Optionally push it to cloud storage if it is a remote dataset.
    """

    check_exists(name, bucket_storage)

    if LuxonisDataset.exists(
        new_name, bucket_storage=BucketStorage.LOCAL
    ) and not Confirm.ask(
        f"Dataset '{new_name}' already exists locally. Overwrite it?"
    ):
        raise typer.Exit

    print(f"Cloning dataset '{name}' to '{new_name}'...")
    dataset = LuxonisDataset(name, bucket_storage=bucket_storage)
    dataset.clone(new_dataset_name=new_name, push_to_cloud=push_to_cloud)
    print(f"[green]Dataset '{name}' successfully cloned to '{new_name}'.")


@app.command()
def merge(
    source_name: Annotated[
        str,
        typer.Argument(
            ...,
            help="Name of the source dataset.",
            autocompletion=complete_dataset_name,
        ),
    ],
    target_name: Annotated[
        str,
        typer.Argument(
            ...,
            help="Name of the target dataset.",
            autocompletion=complete_dataset_name,
        ),
    ],
    new_name: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--new-name",
            "-n",
            help="Name for the new merged dataset. If not provided, will merge into the target dataset.",
            show_default=False,
        ),
    ] = None,
    bucket_storage: BucketStorage = bucket_option,
):
    """Merge two datasets stored in the same type of bucket."""
    check_exists(source_name, bucket_storage)
    check_exists(target_name, bucket_storage)

    inplace = new_name is None
    if inplace and not Confirm.ask(
        f"This will merge dataset '{source_name}' into '{target_name}'. Continue?"
    ):
        raise typer.Exit

    if (
        not inplace
        and LuxonisDataset.exists(new_name, bucket_storage=bucket_storage)
        and not Confirm.ask(
            f"Dataset '{new_name}' already exists in {bucket_storage.value} bucket. Overwrite it?"
        )
    ):
        raise typer.Exit

    source_dataset = LuxonisDataset(source_name, bucket_storage=bucket_storage)
    target_dataset = LuxonisDataset(target_name, bucket_storage=bucket_storage)

    operation = "into" if inplace else "creating new dataset"
    print(
        f"Merging dataset '{source_name}' with '{target_name}', {operation}..."
    )

    _ = target_dataset.merge_with(
        source_dataset, inplace=inplace, new_dataset_name=new_name
    )

    if inplace:
        print(
            f"[green]Dataset '{source_name}' successfully merged into '{target_name}'."
        )
    else:
        print(
            f"[green]Datasets merged successfully into new dataset '{new_name}'."
        )


if __name__ == "__main__":
    app()



from pathlib import Path

import polars as pl
import pytest
from pytest_subtests import SubTests

from luxonis_ml.data import LuxonisLoader, LuxonisParser
from luxonis_ml.enums.enums import DatasetType


@pytest.mark.parametrize("url", ["COCO_people_subset.zip"])
def test_dir_parser(
    dataset_name: str,
    url: str,
    storage_url: str,
    tempdir: Path,
):
    url = f"{storage_url}/{url}"
    dataset = LuxonisParser(
        url,
        dataset_name=dataset_name,
        delete_local=True,
        save_dir=tempdir,
    ).parse()

    metadata = dataset._metadata.model_dump()
    del metadata["tasks"]
    del metadata["skeletons"]
    df = dataset._load_df_offline(raise_when_empty=True)
    anns = (
        df.filter(pl.col("task_type").is_in(["keypoints", "boundingbox"]))
        .select(
            [
                "task_name",
                "class_name",
                "instance_id",
                "task_type",
                "annotation",
            ]
        )
        .to_dict(as_series=False)
    )
    anns = {k: sorted(v) for k, v in anns.items()}

    zip_result = dataset.export(tempdir / "exported")
    zip_path = zip_result[0] if isinstance(zip_result, list) else zip_result
    exported_dataset = LuxonisParser(
        str(zip_path / dataset_name),
        dataset_type=DatasetType.NATIVE,
        dataset_name=dataset_name,
        delete_local=True,
        save_dir=tempdir,
    ).parse()
    imported_metadata = exported_dataset._metadata.model_dump()
    imported_anns = (
        exported_dataset._load_df_offline(raise_when_empty=True)
        .filter(pl.col("task_type").is_in(["keypoints", "boundingbox"]))
        .select(
            [
                "task_name",
                "class_name",
                "instance_id",
                "task_type",
                "annotation",
            ]
        )
        .to_dict(as_series=False)
    )
    imported_anns = {k: sorted(v) for k, v in imported_anns.items()}
    del imported_metadata["tasks"]
    del imported_metadata["skeletons"]
    assert imported_metadata == metadata
    assert imported_anns == anns


@pytest.mark.parametrize("url", ["COCO_people_subset.zip"])
def test_export_edge_cases(
    dataset_name: str,
    storage_url: str,
    tempdir: Path,
    url: str,
    subtests: SubTests,
):
    url = f"{storage_url}/{url}"
    dataset = LuxonisParser(
        url,
        dataset_name=dataset_name,
        delete_local=True,
        save_dir=tempdir,
    ).parse()

    dataset.make_splits(ratios=(1, 0, 0), replace_old_splits=True)

    loader = LuxonisLoader(dataset, view="train")
    original_data = [img for img, _ in loader]

    with subtests.test("Export with max_zip_size_gb=0.003"):
        dataset.export(
            output_path=tempdir / "exported",
            max_partition_size_gb=0.003,
            zip_output=True,
        )
        zip_files = sorted((tempdir / "exported").glob("*.zip"))
        assert len(zip_files) == 2

    with subtests.test(
        "Parse zip dirs into single dataset, having some splits empty"
    ):
        for i, zip_file in enumerate(zip_files):
            dataset = LuxonisParser(
                str(zip_file),
                dataset_name=dataset_name,
                delete_local=(i == 0),
                save_dir=tempdir,
            ).parse()
        dataset.make_splits(ratios=(1, 0, 0), replace_old_splits=True)
        loader = LuxonisLoader(dataset, view="train")
        new_data = [img for img, _ in loader]
        assert len(new_data) == len(original_data)


@pytest.mark.parametrize("url", ["COCO_people_subset.zip"])
def test_export_regular_splits(
    dataset_name: str,
    storage_url: str,
    tempdir: Path,
    url: str,
):
    url = f"{storage_url}/{url}"
    dataset = LuxonisParser(
        url,
        dataset_name=dataset_name,
        delete_local=True,
        save_dir=tempdir,
    ).parse()

    original_splits = dataset.get_splits()
    assert original_splits is not None

    dataset.export(
        output_path=tempdir / "exported",
        max_partition_size_gb=0.003,
        zip_output=True,
    )
    zip_files = sorted((tempdir / "exported").glob("*.zip"))

    assert len(zip_files) == 2

    for i, zip_file in enumerate(zip_files):
        dataset = LuxonisParser(
            str(zip_file),
            dataset_name=dataset_name,
            delete_local=(i == 0),
            save_dir=tempdir,
        ).parse()

    new_splits = dataset.get_splits()
    assert new_splits is not None
    assert len(new_splits) == len(original_splits)

    for split in original_splits:
        original_split = sorted(original_splits[split])
        new_split = sorted(new_splits[split])
        assert original_split == new_split



from abc import ABC, abstractmethod
from typing import Iterator, Type

from luxonis_ml.typing import LoaderOutput
from luxonis_ml.utils import AutoRegisterMeta, Registry

LOADERS_REGISTRY: Registry[Type["BaseLoader"]] = Registry(name="loaders")


class BaseLoader(
    ABC, metaclass=AutoRegisterMeta, registry=LOADERS_REGISTRY, register=False
):
    """Base abstract loader class.

    Enforces the L{LuxonisLoaderOutput} output label structure.
    """

    @abstractmethod
    def __len__(self) -> int:
        """Returns the length of the dataset.

        @rtype: int
        @return: Length of the dataset.
        """
        ...

    @abstractmethod
    def __getitem__(self, idx: int) -> LoaderOutput:
        """Loads sample from dataset.

        @type idx: int
        @param idx: Index of the sample to load.
        @rtype: L{LuxonisLoaderOutput}
        @return: Sample's data in C{LuxonisLoaderOutput} format.
        """
        ...

    def __iter__(self) -> Iterator[LoaderOutput]:
        """Iterates over the dataset.

        @rtype: Iterator[L{LuxonisLoaderOutput}]
        @return: Iterator over the dataset.
        """
        for i in range(len(self)):
            yield self[i]



from pathlib import Path
from typing import Dict, Set, Tuple, Union

import cv2
import numpy as np

from luxonis_ml.data import LuxonisLoader
from luxonis_ml.data.datasets.base_dataset import DatasetIterator
from luxonis_ml.data.datasets.luxonis_dataset import LuxonisDataset
from luxonis_ml.data.utils.enums import BucketStorage


def gather_tasks(dataset: LuxonisDataset) -> Set[str]:
    return {
        f"{task_name}/{task_type}"
        for task_name, task_types in dataset.get_tasks().items()
        for task_type in task_types
    }


def create_image(i: int, dir: Path) -> Path:
    path = dir / f"img_{i}.jpg"
    if not path.exists():
        img = np.zeros((512, 512, 3), dtype=np.uint8)
        img[0:10, 0:10] = np.random.randint(
            0, 255, (10, 10, 3), dtype=np.uint8
        )
        cv2.imwrite(str(path), img)
    return path


def get_loader_output(loader: LuxonisLoader) -> Set[str]:
    all_labels = set()
    for _, labels in loader:
        all_labels.update(labels.keys())
    return all_labels


def create_dataset(
    dataset_name: str,
    generator: DatasetIterator,
    bucket_storage: BucketStorage = BucketStorage.LOCAL,
    *,
    splits: Union[bool, Dict[str, float], Tuple] = True,
    delete_local: bool = True,
    delete_remote: bool = True,
    **kwargs,
) -> LuxonisDataset:
    dataset = LuxonisDataset(
        dataset_name,
        delete_local=delete_local,
        delete_remote=delete_remote,
        bucket_storage=bucket_storage,
        **kwargs,
    ).add(generator)
    if splits is True:
        dataset.make_splits()
    elif splits:
        dataset.make_splits(splits)
    return dataset



import random
from contextlib import contextmanager
from pathlib import Path
from typing import List

import numpy as np

from luxonis_ml.data import (
    BucketStorage,
    LuxonisDataset,
    LuxonisLoader,
    LuxonisParser,
)
from luxonis_ml.data.datasets.base_dataset import DatasetIterator
from luxonis_ml.enums import DatasetType
from luxonis_ml.typing import Params

from .utils import create_image


@contextmanager
def set_seed(seed: int):
    np_state = np.random.get_state()
    random_state = random.getstate()
    np.random.seed(seed)
    random.seed(seed)
    yield
    np.random.set_state(np_state)
    random.setstate(random_state)


def test_edge_cases(tempdir: Path):
    keypoints_before = np.array(
        [
            [
                0.41100962,
                0.63641827,
                2.0,
                0.49552885,
                0.61091346,
                2.0,
                0.50201923,
                0.63555288,
                2.0,
                0.41879808,
                0.66302885,
                2.0,
            ],
            [
                0.86091346,
                0.76336538,
                2.0,
                0.99495192,
                0.71382212,
                2.0,
                0.97978365,
                0.75096154,
                0.0,
                0.87086538,
                0.7909375,
                2.0,
            ],
            [
                0.77483173,
                0.23288462,
                2.0,
                0.90862981,
                0.16939904,
                2.0,
                0.91990385,
                0.19620192,
                2.0,
                0.78622596,
                0.26213942,
                2.0,
            ],
            [
                0.07838942,
                0.40788462,
                2.0,
                0.12990385,
                0.59675481,
                2.0,
                0.10711538,
                0.60266827,
                2.0,
                0.05271635,
                0.41540865,
                2.0,
            ],
            [
                0.03276442,
                0.38091346,
                2.0,
                0.08644231,
                0.5690625,
                2.0,
                0.06149038,
                0.57569712,
                2.0,
                0.00709135,
                0.3884375,
                2.0,
            ],
            [
                0.0,
                0.39269231,
                0.0,
                0.01310096,
                0.5846875,
                2.0,
                0.0,
                0.58747596,
                0.0,
                0.0,
                0.40019231,
                0.0,
            ],
            [
                0.06036058,
                0.23735577,
                2.0,
                0.0,
                0.25528846,
                0.0,
                0.0,
                0.23185096,
                0.0,
                0.05483173,
                0.21627404,
                2.0,
            ],
            [
                0.09098558,
                0.72336538,
                2.0,
                0.14283654,
                0.90288462,
                2.0,
                0.12247596,
                0.91052885,
                2.0,
                0.06740385,
                0.72949519,
                2.0,
            ],
            [
                0.04668269,
                0.18875,
                2.0,
                0.0,
                0.20668269,
                0.0,
                0.0,
                0.18322115,
                0.0,
                0.04115385,
                0.16766827,
                2.0,
            ],
            [
                0.02704327,
                0.14060096,
                2.0,
                0.0,
                0.15853365,
                0.0,
                0.0,
                0.13507212,
                0.0,
                0.02149038,
                0.11951923,
                0.0,
            ],
            [
                0.41819712,
                0.90716346,
                2.0,
                0.45608173,
                0.89471154,
                2.0,
                0.47170673,
                0.94043269,
                2.0,
                0.43572115,
                0.95435096,
                2.0,
            ],
        ]
    )

    bboxes_before = np.array(
        [
            [0.0, 0.41100962, 0.61091346, 0.09100962, 0.05211538],
            [0.0, 0.86091346, 0.71382212, 0.13403846, 0.07711538],
            [0.0, 0.77483173, 0.16939904, 0.14507212, 0.09274038],
            [0.0, 0.05271635, 0.40788462, 0.0771875, 0.19478365],
            [0.0, 0.00709135, 0.38091346, 0.07935096, 0.19478365],
            [0.0, 0.0, 0.39269231, 0.01310096, 0.19478365],
            [0.0, 0.0, 0.21627404, 0.06036058, 0.03901442],
            [0.0, 0.06740385, 0.72336538, 0.07543269, 0.18716346],
            [0.0, 0.0, 0.16766827, 0.04668269, 0.03901442],
            [0.0, 0.0, 0.11951923, 0.02704327, 0.03901442],
            [0.0, 0.41819712, 0.89471154, 0.05350962, 0.05963942],
        ]
    )

    def generator() -> DatasetIterator:
        num_samples = len(keypoints_before)
        for i in range(num_samples):
            img = create_image(i, tempdir)
            keypoints_list = keypoints_before[i].tolist()
            bbox_list = bboxes_before[i].tolist()

            keypoints_split = [
                [x, y, int(visibility)]
                for x, y, visibility in (
                    keypoints_list[j : j + 3]
                    for j in range(0, len(keypoints_list), 3)
                )
            ]

            yield {
                "file": img,
                "annotation": {
                    "class": "dog",
                    "boundingbox": {
                        "x": bbox_list[1],
                        "y": bbox_list[2],
                        "w": bbox_list[3],
                        "h": bbox_list[4],
                    },
                    "keypoints": {"keypoints": keypoints_split},
                },
            }

    dataset = LuxonisDataset(
        "test_edge_cases",
        delete_local=True,
        delete_remote=True,
        bucket_storage=BucketStorage.LOCAL,
    ).add(generator())

    dataset.make_splits(ratios=(1, 0, 0))

    augmentation_config: List[Params] = [
        {
            "name": "Mosaic4",
            "params": {"p": 1, "out_width": 512, "out_height": 512},
        },
        {
            "name": "Rotate",
            "params": {
                "limit": 10,
                "p": 1,
                "border_mode": 0,
                "value": [0, 0, 0],
            },
        },
        {
            "name": "Perspective",
            "params": {
                "scale": [0.02, 0.05],
                "keep_size": True,
                "pad_mode": 0,
                "pad_val": 0,
                "mask_pad_val": 0,
                "fit_output": False,
                "interpolation": 1,
                "always_apply": False,
                "p": 1,
            },
        },
        {
            "name": "Affine",
            "params": {
                "scale": 1.0,
                "translate_percent": 0.0,
                "rotate": 0,
                "shear": 5,
                "interpolation": 1,
                "mask_interpolation": 0,
                "cval": 0,
                "cval_mask": 0,
                "mode": 0,
                "fit_output": False,
                "keep_ratio": False,
                "rotate_method": "largest_box",
                "always_apply": False,
                "p": 1,
            },
        },
    ]

    loader = LuxonisLoader(
        dataset,
        view="train",
        augmentation_config=augmentation_config,
        height=512,
        width=512,
    )
    for _ in range(20):
        for batch in loader:
            keypoints = batch[1]["/keypoints"]
            bboxes = batch[1]["/boundingbox"]
            keypoints = keypoints.reshape(-1, 3)
            for kp in keypoints:
                x, y, v = kp
                assert not (x == 0 and y == 0 and v in {1, 2}), (
                    f"Invalid keypoint detected: {kp}"
                )
                assert 0 <= x <= 1, f"Keypoint x out of bounds: {kp}"
                assert 0 <= y <= 1, f"Keypoint y out of bounds: {kp}"

            for bbox in bboxes:
                _, x_min, y_min, width, height = bbox

                x_max = x_min + width
                y_max = y_min + height

                assert 0 <= x_min <= 1, f"BBox x_min out of bounds: {bbox}"
                assert 0 <= y_min <= 1, f"BBox y_min out of bounds: {bbox}"
                assert 0 <= x_max <= 1, f"BBox x_max out of bounds: {bbox}"
                assert 0 <= y_max <= 1, f"BBox y_max out of bounds: {bbox}"

                bbox_area = width * height
                assert bbox_area >= 0.0004, (
                    f"BBox area too small: {bbox}, area={bbox_area}"
                )


def test_dataset_reproducibility(storage_url: str, tempdir: Path):
    def create_loader(storage_url: str, tempdir: Path) -> LuxonisLoader:
        with set_seed(42):
            dataset = LuxonisParser(
                f"{storage_url}/COCO_people_subset.zip",
                dataset_name="_augmentation_reproducibility",
                save_dir=tempdir,
                dataset_type=DatasetType.COCO,
                delete_local=True,
            ).parse()
            return LuxonisLoader(
                dataset,
                height=512,
                width=512,
                view="train",
            )

    loader1 = create_loader(storage_url, tempdir)
    run1 = [ann for _, ann in loader1]

    loader2 = create_loader(storage_url, tempdir)
    run2 = [ann for _, ann in loader2]

    assert all(
        a1.keys() == a2.keys()
        and all(
            np.array_equal(a1[k], a2[k])
            if isinstance(a1[k], np.ndarray)
            else a1[k] == a2[k]
            for k in a1
        )
        for a1, a2 in zip(run1, run2)
    )


def test_filter_by_task_name(tempdir: Path):
    def generator() -> DatasetIterator:
        for i in range(6):
            yield {
                "file": create_image(i, tempdir),
                "task_name": "person_cls" if i % 2 == 0 else "car_cls",
                "annotation": {"class": "person" if i % 2 == 0 else "car"},
            }

    dataset = LuxonisDataset(
        "test_filter_by_task_name",
        delete_local=True,
        bucket_storage=BucketStorage.LOCAL,
    )
    dataset.add(generator())
    dataset.make_splits(splits=(1, 0, 0))

    for task_name in ["person_cls", "car_cls"]:
        loader = LuxonisLoader(
            dataset, view="train", filter_task_names=[task_name]
        )
        assert len(loader) == 3



from abc import ABC, abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from luxonis_ml.data import BaseDataset, DatasetIterator
from luxonis_ml.data.datasets.annotation import DatasetRecord
from luxonis_ml.enums.enums import DatasetType

ParserOutput = Tuple[DatasetIterator, Dict[str, Dict], List[Path]]


class BaseParser(ABC):
    def __init__(
        self,
        dataset: BaseDataset,
        dataset_type: DatasetType,
        task_name: Optional[Union[str, Dict[str, str]]],
    ):
        """
        @type dataset: BaseDataset
        @param dataset: Dataset to add the parsed data to.
        @type dataset_type: DatasetType
        @param dataset_type: Type of the dataset.
        @type task_name: Optional[Union[str, Dict[str, str]]]
        @param task_name: Optional task name(s) for the dataset.
            Can be either a single string, in which case all the records
            added to the dataset will use this value as `task_name`, or
            a dictionary with class names as keys and task names as values.
            In the latter case, the task name for a record with a given
            class name will be taken from the dictionary.
        """
        self.dataset = dataset
        self.dataset_type = dataset_type
        if isinstance(task_name, str):
            self.task_name = defaultdict(lambda: task_name)
        else:
            self.task_name = task_name

    @staticmethod
    @abstractmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        """Validates if a split subdirectory is in an expected format.
        If so, returns kwargs to pass to L{from_split} method.

        @type split_path: Path
        @param split_path: Path to split directory.
        @rtype: Optional[Dict[str, Any]]
        @return: Dictionary with kwargs to pass to L{from_split} method
            or C{None} if the split is not in the expected format.
        """
        ...

    @staticmethod
    @abstractmethod
    def validate(dataset_dir: Path) -> bool:
        """Validates if the dataset is in an expected format.

        @type dataset_dir: Path
        @param dataset_dir: Path to source dataset directory.
        @rtype: bool
        @return: If the dataset is in the expected format.
        """
        ...

    @abstractmethod
    def from_dir(
        self, dataset_dir: Path, **kwargs
    ) -> Tuple[List[str], List[str], List[str]]:
        """Parses all present data to L{LuxonisDataset} format.

        @type dataset_dir: str
        @param dataset_dir: Path to source dataset directory.
        @type kwargs: Any
        @param kwargs: Additional arguments for a specific parser
            implementation.
        @rtype: Tuple[List[str], List[str], List[str]]
        @return: Tuple with added images for C{train}, C{val} and
            C{test} splits.
        """
        ...

    @abstractmethod
    def from_split(self, **kwargs) -> ParserOutput:
        """Parses a data in a split subdirectory to L{LuxonisDataset}
        format.

        @type kwargs: Dict[str, Any]
        @param kwargs: Additional kwargs for specific parser implementation.
            Should work together with L{validate_split} method like:

                >>> from_split(**validate_split(split_path))

        @rtype: ParserOutput
        @return: C{LDF} generator, list of class names,
            skeleton dictionary for keypoints and list of added images.
        """
        ...

    def _parse_split(self, **kwargs) -> List[Path]:
        """Parses data in a split subdirectory.

        @type kwargs: Dict[str, Any]
        @param kwargs: Additional kwargs for specific parser
            implementation.
        @rtype: List[str]
        @return: List of added images.
        """
        generator, skeletons, added_images = self.from_split(**kwargs)
        self.dataset.add(self._wrap_generator(generator))
        if skeletons:
            for skeleton in skeletons.values():
                self.dataset.set_skeletons(
                    skeleton.get("labels"),
                    skeleton.get("edges"),
                )
        return added_images

    def parse_split(
        self,
        split: Optional[str] = None,
        random_split: bool = False,
        split_ratios: Optional[Dict[str, float]] = None,
        **kwargs,
    ) -> BaseDataset:
        """Parses data in a split subdirectory to L{LuxonisDataset}
        format.

        @type split: Optional[str]
        @param split: As what split the data will be added to LDF. If
            set, C{split_ratios} and C{random_split} are ignored.
        @type random_split: bool
        @param random_split: If random splits should be made. If
            C{True}, C{split_ratios} are used.
        @type split_ratios: Optional[Tuple[float, float, float]]
        @param split_ratios: Ratios for random splits. Only used if
            C{random_split} is C{True}. Defaults to C{(0.8, 0.1, 0.1)}.
        @type kwargs: Dict[str, Any]
        @param kwargs: Additional C{kwargs} for specific parser
            implementation.
        @rtype: LuxonisDataset
        @return: C{LDF} with all the images and annotations parsed.
        """
        added_images = self._parse_split(**kwargs)
        if split is not None:
            self.dataset.make_splits({split: added_images})
        elif random_split:
            self.dataset.make_splits(split_ratios)
        return self.dataset

    def parse_dir(self, dataset_dir: Path, **kwargs) -> BaseDataset:
        """Parses entire dataset directory to L{LuxonisDataset} format.

        @type dataset_dir: str
        @param dataset_dir: Path to source dataset directory.
        @type kwargs: Dict[str, Any]
        @param kwargs: Additional C{kwargs} for specific parser
            implementation.
        @rtype: LuxonisDataset
        @return: C{LDF} with all the images and annotations parsed.
        """
        train, val, test = self.from_dir(dataset_dir, **kwargs)

        self.dataset.make_splits({"train": train, "val": val, "test": test})
        return self.dataset

    @staticmethod
    def _get_added_images(generator: DatasetIterator) -> List[Path]:
        """Returns list of unique images added by the generator
        function.

        @type generator: L{DatasetGenerator}
        @param generator: Generator function
        @rtype: List[str]
        @return: List of added images by generator function
        """
        return list(
            {
                Path(item["file"] if isinstance(item, dict) else item.file)
                for item in generator
            }
        )

    @staticmethod
    def _compare_stem_files(
        list1: Iterable[Path], list2: Iterable[Path]
    ) -> bool:
        """Compares sets of files by their stem.

        Example:

            >>> BaseParser._compare_stem_files([Path("a.jpg"), Path("b.jpg")],
            ...                                [Path("a.xml"), Path("b.xml")])
            True
            >>> BaseParser._compare_stem_files([Path("a.jpg")], [Path("b.txt")])
            False

        @type list1: Iterable[Path]
        @param list1: First list of files
        @type list2: Iterable[Path]
        @param list2: Second list of files
        @rtype: bool
        @return: If the two sets of files are equal when compared by their stems.
            If the sets are empty, returns C{False}.
        """
        set1 = {Path(f).stem for f in list1}
        set2 = {Path(f).stem for f in list2}
        return len(set1) > 0 and set1 == set2

    @staticmethod
    def _list_images(image_dir: Path) -> List[Path]:
        """Returns list of all images in the directory supported by
        opencv.

        @type image_dir: Path
        @param image_dir: Path to directory with images
        @rtype: List[Path]
        @return: List of images in the directory
        """
        cv2_supported_image_formats = {
            ".bmp",
            ".dib",
            ".jpeg",
            ".jpg",
            ".jpe",
            ".jp2",
            ".png",
            ".WebP",
            ".webp",
            ".pbm",
            ".pgm",
            ".ppm",
            ".pxm",
            ".pnm",
            ".sr",
            ".ras",
            ".tiff",
            ".tif",
            ".exr",
            ".hdr",
            ".pic",
        }
        return [
            img
            for img in image_dir.glob("*")
            if img.suffix in cv2_supported_image_formats
        ]

    def _wrap_generator(self, generator: DatasetIterator) -> DatasetIterator:
        """Adds task to the generator.

        @type generator: DatasetIterator
        @param generator: Generator function
        @rtype: DatasetIterator
        @return: Generator function with added task
        """

        for item in generator:
            if isinstance(item, dict):
                item = DatasetRecord(**item)

            if self.task_name is not None:
                if item.annotation is None:
                    for task_name in set(self.task_name.values()):
                        yield item.model_copy(
                            update={"task_name": task_name}, deep=True
                        )
                else:
                    class_name = item.annotation.class_name
                    if class_name is not None:
                        try:
                            task_name = self.task_name[class_name]
                        except KeyError:
                            raise ValueError(
                                f"Class '{class_name}' not found in task names."
                            ) from None

                        item.task_name = self.task_name[class_name]
                    yield item
            else:
                yield item



import shutil
from pathlib import Path
from types import ModuleType
from typing import Literal, Optional

import polars as pl
from typing_extensions import overload

from luxonis_ml.typing import PathType, PosixPathType
from luxonis_ml.utils.filesystem import LuxonisFileSystem


@overload
def get_file(
    fs: LuxonisFileSystem,
    remote_path: PosixPathType,
    local_path: PathType,
    mlflow_instance: Optional[ModuleType] = ...,
    default: Literal[None] = ...,
) -> Optional[Path]: ...


@overload
def get_file(
    fs: LuxonisFileSystem,
    remote_path: PosixPathType,
    local_path: PathType,
    mlflow_instance: Optional[ModuleType] = ...,
    default: PathType = ...,
) -> Path: ...


def get_file(
    fs: LuxonisFileSystem,
    remote_path: PosixPathType,
    local_path: PathType,
    mlflow_instance: Optional[ModuleType] = None,
    default: Optional[PathType] = None,
) -> Optional[Path]:
    try:
        return fs.get_file(remote_path, local_path, mlflow_instance)
    except shutil.SameFileError:
        return Path(local_path, Path(remote_path).name)
    except Exception:
        return Path(default) if default is not None else None


@overload
def find_filepath_uuid(
    filepath: PathType,
    index: Optional[pl.DataFrame],
    *,
    raise_on_missing: Literal[False] = ...,
) -> Optional[str]: ...


@overload
def find_filepath_uuid(
    filepath: PathType,
    index: Optional[pl.DataFrame],
    *,
    raise_on_missing: Literal[True] = ...,
) -> str: ...


def find_filepath_uuid(
    filepath: PathType,
    index: Optional[pl.DataFrame],
    *,
    raise_on_missing: bool = False,
) -> Optional[str]:
    if index is None:
        return None

    abs_path = str(Path(filepath).absolute().resolve())
    matched = index.filter(pl.col("original_filepath") == abs_path)

    if len(matched):
        return next(iter(matched.select("uuid")))[0]
    if raise_on_missing:
        raise ValueError(f"File {abs_path} not found in index")
    return None


@overload
def get_dir(
    fs: LuxonisFileSystem,
    remote_path: PosixPathType,
    local_dir: PathType,
    mlflow_instance: Optional[ModuleType] = ...,
    *,
    default: Literal[None] = None,
) -> Optional[Path]: ...


@overload
def get_dir(
    fs: LuxonisFileSystem,
    remote_path: PosixPathType,
    local_dir: PathType,
    mlflow_instance: Optional[ModuleType] = ...,
    *,
    default: Path = ...,
) -> Path: ...


def get_dir(
    fs: LuxonisFileSystem,
    remote_path: PosixPathType,
    local_dir: PathType,
    mlflow_instance: Optional[ModuleType] = None,
    *,
    default: Optional[PathType] = None,
) -> Optional[Path]:
    try:
        return fs.get_dir(remote_path, local_dir, mlflow_instance)
    except shutil.SameFileError:
        return Path(local_dir, Path(remote_path).name)
    except Exception:
        return Path(default) if default is not None else None



from collections import defaultdict
from typing import (
    Any,
    Dict,
    Final,
    List,
    Literal,
    Optional,
    Set,
    Union,
    overload,
)

import polars as pl
from typing_extensions import TypedDict

from .metadata import Metadata, Skeletons

LDF_1_0_0_TASKS: Final[Set[str]] = {
    "classification",
    "segmentation",
    "boundingbox",
    "keypoints",
    "array",
}

LDF_1_0_0_TASK_TYPES: Final[Dict[str, str]] = {
    "BBoxAnnotation": "boundingbox",
    "ClassificationAnnotation": "classification",
    "PolylineSegmentationAnnotation": "segmentation",
    "RLESegmentationAnnotation": "segmentation",
    "MaskSegmentationAnnotation": "segmentation",
    "KeypointAnnotation": "keypoints",
    "ArrayAnnotation": "array",
}


class LDF_1_0_0_MetadataDict(TypedDict):
    source: Dict[str, Any]
    ldf_version: str
    classes: Dict[str, List[str]]
    tasks: Dict[str, List[str]]
    skeletons: Dict[str, Skeletons]
    categorical_encodings: Dict[str, Dict[str, int]]
    metadata_types: Dict[str, Literal["float", "int", "str", "Category"]]


@overload
def migrate_dataframe(df: pl.LazyFrame) -> pl.LazyFrame: ...


@overload
def migrate_dataframe(df: pl.DataFrame) -> pl.DataFrame: ...


def migrate_dataframe(
    df: Union[pl.LazyFrame, pl.DataFrame],
) -> Union[pl.LazyFrame, pl.DataFrame]:  # pragma: no cover
    return (
        df.rename({"class": "class_name"})
        .with_columns(
            pl.when(pl.col("task").is_in(LDF_1_0_0_TASKS))
            .then(pl.lit("detection"))
            .otherwise(pl.col("task"))
            .alias("task_name")
        )
        .with_columns(
            pl.when(pl.col("type") == "BBoxAnnotation")
            .then(pl.lit("boundingbox"))
            .when(pl.col("type") == "ClassificationAnnotation")
            .then(pl.lit("classification"))
            .when(
                pl.col("type").is_in(
                    [
                        "PolylineSegmentationAnnotation",
                        "RLESegmentationAnnotation",
                        "MaskSegmentationAnnotation",
                    ]
                )
            )
            .then(pl.lit("segmentation"))
            .when(pl.col("type") == "KeypointAnnotation")
            .then(pl.lit("keypoints"))
            .when(pl.col("type") == "ArrayAnnotation")
            .then(pl.lit("array"))
            .otherwise(pl.col("type"))
            .alias("task_type")
        )
        .with_columns(pl.lit("image").alias("source_name"))
        .select(
            [
                "file",
                "source_name",
                "task_name",
                "class_name",
                "instance_id",
                "task_type",
                "annotation",
                "uuid",
            ]
        )
    )


def migrate_metadata(
    metadata: LDF_1_0_0_MetadataDict, df: Optional[pl.LazyFrame]
) -> Metadata:  # pragma: no cover
    new_metadata = {}
    old_classes = metadata["classes"]
    if set(old_classes.keys()) <= LDF_1_0_0_TASKS:
        old_class_names = next(iter(old_classes.values()))
        new_metadata["classes"] = {
            "detection": {
                class_name: i for i, class_name in enumerate(old_class_names)
            }
        }
        new_metadata["tasks"] = {"detection": list(old_classes.keys())}
    else:
        if df is None:
            raise ValueError("Cannot migrate when the dataset is empty")
        tasks_df = df.select(["task", "type"]).unique().collect()
        new_classes = defaultdict(dict)
        tasks = defaultdict(list)
        for task_name, task_type in tasks_df.iter_rows():
            new_task_name = task_name
            if task_name in LDF_1_0_0_TASKS:
                new_task_name = "detection"
            tasks[new_task_name].append(LDF_1_0_0_TASK_TYPES[task_type])
            old_class_names = old_classes.get(task_name, [])
            new_classes[new_task_name].update(
                {class_name: i for i, class_name in enumerate(old_class_names)}
            )
        new_metadata["classes"] = dict(new_classes)
        new_metadata["tasks"] = dict(tasks)

    metadata.update(new_metadata)
    return Metadata(**metadata)  # type: ignore



import json
import tarfile
from io import BytesIO
from pathlib import Path
from typing import List, Literal, Tuple

from luxonis_ml.typing import PathType

from .config import Config


class ArchiveGenerator:
    """Generator of abstracted NN archive (.tar) files containing config
    and model files (executables).

    @type archive_name: str
    @ivar archive_name: Desired archive file name.
    @type save_path: str
    @ivar save_path: Path to where we want to save the archive file.
    @type cfg_dict: dict
    @ivar cfg_dict: Archive configuration dict.
    @type executables_paths: list
    @ivar executables_paths: Paths to relevant model executables.
    @type compression: str
    @ivar compression: Type of archive file compression ("xz" for LZMA,
        "gz" for gzip, or "bz2" for bzip2 compression).
    """

    def __init__(
        self,
        archive_name: str,
        save_path: PathType,
        cfg_dict: dict,
        executables_paths: List[PathType],
        compression: Literal["xz", "gz", "bz2"] = "xz",
    ):
        self.save_path = Path(save_path)
        self.executables_paths = executables_paths

        if compression not in ["xz", "gz", "bz2"]:
            raise ValueError(
                "Invalid compression type. Must be one of 'xz', 'gz', 'bz2'."
            )
        self.compression = compression

        self.archive_name = (
            archive_name
            if archive_name.endswith(f".tar.{self.compression}")
            else f"{archive_name}.tar.{self.compression}"
        )

        self.cfg = Config(  # pydantic config check
            config_version=cfg_dict["config_version"], model=cfg_dict["model"]
        )

    def make_archive(self) -> Path:
        """Run NN archive (.tar) file generation."""

        # create an in-memory file-like config object
        json_data, json_buffer = self._make_json()

        # construct .tar archive
        archive_path = self.save_path / self.archive_name
        with tarfile.open(
            archive_path,
            f"w:{self.compression}",  # type: ignore
        ) as tar:
            # add executables
            for executable_path in map(Path, self.executables_paths):
                tar.add(executable_path, arcname=executable_path.name)
            # add config JSON
            tarinfo = tarfile.TarInfo(name="config.json")
            tarinfo.size = len(json_data)
            json_buffer.seek(0)  # reset the buffer to the beginning
            tar.addfile(tarinfo, json_buffer)

        return archive_path

    def _make_json(self) -> Tuple[bytes, BytesIO]:
        """Create an in-memory config data file-like object."""

        # read-in config data as dict
        data = json.loads(self.cfg.model_dump_json())

        # create an in-memory file-like object
        json_buffer = BytesIO()

        # encode the dictionary as bytes and write it to the in-memory file
        json_data = json.dumps(data, indent=4).encode("utf-8")
        json_buffer.write(json_data)

        return json_data, json_buffer



from enum import Enum


class DatasetType(str, Enum):
    LDF = "ldf"
    COCO = "coco"
    VOC = "voc"
    DARKNET = "darknet"
    YOLOV6 = "yolov6"
    YOLOV4 = "yolov4"
    CREATEML = "createml"
    TFCSV = "tfcsv"
    CLSDIR = "clsdir"
    SEGMASK = "segmask"
    SOLO = "solo"
    NATIVE = "native"



from typing import Iterator, Set, Tuple

import numpy as np

from luxonis_ml.data.utils import (
    get_task_name,
    get_task_type,
    split_task,
    task_is_metadata,
    task_type_iterator,
)


def test_task_is_metadata():
    assert task_is_metadata("a/b/metadata/c")
    assert task_is_metadata("metadata/c")
    assert not task_is_metadata("a/b/c")


def test_split_task():
    assert split_task("a/b") == ("a", "b")
    assert split_task("a") == ("", "a")
    assert split_task("a/b/c/d/e") == ("a", "b/c/d/e")


def test_get_task_name():
    assert get_task_name("a/b") == "a"
    assert get_task_name("a") == "a"
    assert get_task_name("a/b/c") == "a"


def test_get_task_type():
    assert get_task_type("a/b") == "b"
    assert get_task_type("a") == "a"
    assert get_task_type("a/b/c") == "c"
    assert get_task_type("a/b/metadata/c") == "metadata/c"
    assert get_task_type("metadata/c") == "metadata/c"


def test_task_type_iterator():
    labels = {
        "task/segmentation": np.array([]),
        "task/metadata/text": np.array([]),
        "task/instance_segmentation": np.array([]),
        "task/subtask/boundingbox": np.array([]),
        "task/subtask/segmentation": np.array([]),
    }

    def compare(
        iterator: Iterator[Tuple[str, np.ndarray]], expected: Set[str]
    ) -> None:
        tasks = set()
        for task, _ in iterator:
            tasks.add(task)
        assert tasks == expected

    compare(
        task_type_iterator(labels, "segmentation"),
        {"task/segmentation", "task/subtask/segmentation"},
    )
    compare(
        task_type_iterator(labels, "boundingbox"), {"task/subtask/boundingbox"}
    )
    compare(
        task_type_iterator(labels, "instance_segmentation"),
        {"task/instance_segmentation"},
    )
    compare(task_type_iterator(labels, "keypoints"), set())
